"""
Chunking modules for refinire-rag
"""

from .chunker import Chunker, ChunkingConfig
from .token_chunker import TokenBasedChunker
from .sentence_chunker import SentenceAwareChunker

__all__ = [
    "Chunker",
    "ChunkingConfig",
    "TokenBasedChunker",
    "SentenceAwareChunker",
]


"""
Tests for GraphBuilder processor

Tests the knowledge graph construction and relationship extraction functionality.
"""

import os
import pytest
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch, MagicMock

from refinire_rag.processing.graph_builder import GraphBuilder, GraphBuilderConfig
from refinire_rag.models.document import Document


class TestGraphBuilderConfig:
    """Test GraphBuilderConfig configuration class"""

    def test_default_config(self):
        """Test default configuration values"""
        config = GraphBuilderConfig()
        
        # Paths should be converted to absolute
        assert config.graph_file_path.endswith("domain_knowledge_graph.md")
        assert config.dictionary_file_path.endswith("domain_dictionary.md")
        assert os.path.isabs(config.graph_file_path)
        assert os.path.isabs(config.dictionary_file_path)
        assert config.backup_graph is True
        assert config.llm_model is not None  # Should be set by __post_init__
        assert config.llm_temperature == 0.3
        assert config.max_tokens == 3000
        assert config.focus_on_important_relationships is True
        assert config.extract_hierarchical_relationships is True
        assert config.extract_causal_relationships is True

    def test_custom_config(self):
        """Test custom configuration values"""
        config = GraphBuilderConfig(
            graph_file_path="/custom/path/graph.md",
            dictionary_file_path="/custom/path/dict.md",
            llm_model="gpt-4",
            llm_temperature=0.5,
            focus_on_important_relationships=False
        )
        
        assert config.graph_file_path == "/custom/path/graph.md"
        assert config.dictionary_file_path == "/custom/path/dict.md"
        assert config.llm_model == "gpt-4"
        assert config.llm_temperature == 0.5
        assert config.focus_on_important_relationships is False

    @patch.dict(os.environ, {"REFINIRE_RAG_LLM_MODEL": "test-model"})
    def test_env_model_config(self):
        """Test that environment variable is used for model selection"""
        config = GraphBuilderConfig()
        assert config.llm_model == "test-model"

    def test_absolute_path_conversion(self):
        """Test that relative paths are converted to absolute"""
        config = GraphBuilderConfig(
            graph_file_path="relative/graph.md",
            dictionary_file_path="relative/dict.md"
        )
        assert os.path.isabs(config.graph_file_path)
        assert os.path.isabs(config.dictionary_file_path)


class TestGraphBuilder:
    """Test GraphBuilder processor"""

    @pytest.fixture
    def temp_dir(self):
        """Create temporary directory for test files"""
        with tempfile.TemporaryDirectory() as tmpdir:
            yield Path(tmpdir)

    @pytest.fixture
    def sample_document(self):
        """Create sample document for testing"""
        return Document(
            id="test_doc_1",
            content="Machine learning algorithms process data to train neural networks. "
                   "The API connects to the database which stores user information. "
                   "Data preprocessing improves model accuracy.",
            metadata={"source": "test"}
        )

    @pytest.fixture
    def config_with_temp_files(self, temp_dir):
        """Create config with temporary files"""
        graph_file = temp_dir / "test_graph.md"
        dict_file = temp_dir / "test_dict.md"
        return GraphBuilderConfig(
            graph_file_path=str(graph_file),
            dictionary_file_path=str(dict_file),
            llm_model="gpt-4o-mini"
        )

    @pytest.fixture
    def sample_dictionary_content(self):
        """Create sample dictionary content"""
        return """# ドメイン用語辞書

## 専門用語
- machine learning: 機械学習
- neural networks: ニューラルネットワーク
- API: Application Programming Interface

## 技術概念
- data preprocessing: データ前処理
- model accuracy: モデル精度
"""

    def test_initialization_without_refinire(self, config_with_temp_files):
        """Test initialization when Refinire is not available"""
        with patch('refinire_rag.processing.graph_builder.LLMPipeline', None):
            builder = GraphBuilder(config_with_temp_files)
            assert builder._llm_pipeline is None
            assert builder.config == config_with_temp_files

    @patch('refinire_rag.processing.graph_builder.LLMPipeline')
    def test_initialization_with_refinire(self, mock_llm_pipeline, config_with_temp_files):
        """Test initialization when Refinire is available"""
        mock_pipeline_instance = Mock()
        mock_llm_pipeline.return_value = mock_pipeline_instance
        
        builder = GraphBuilder(config_with_temp_files)
        
        mock_llm_pipeline.assert_called_once_with(
            name="graph_builder",
            generation_instructions="You are a knowledge graph expert that extracts important relationships from documents.",
            model=config_with_temp_files.llm_model
        )
        assert builder._llm_pipeline == mock_pipeline_instance

    @patch('refinire_rag.processing.graph_builder.LLMPipeline')
    def test_initialization_refinire_error(self, mock_llm_pipeline, config_with_temp_files):
        """Test initialization when Refinire initialization fails"""
        mock_llm_pipeline.side_effect = Exception("API error")
        
        builder = GraphBuilder(config_with_temp_files)
        assert builder._llm_pipeline is None

    def test_empty_graph_template(self, config_with_temp_files):
        """Test creation of empty graph template"""
        builder = GraphBuilder(config_with_temp_files)
        template = builder._create_empty_graph_template()
        
        assert "# ドメイン知識グラフ" in template
        assert "## エンティティ関係" in template
        assert "### 主要概念" in template
        assert "### 技術関係" in template

    def test_process_without_llm(self, config_with_temp_files, sample_document):
        """Test processing when LLM is not available (mock data mode)"""
        builder = GraphBuilder(config_with_temp_files)
        builder._llm_pipeline = None
        
        result = builder.process(sample_document)
        
        assert len(result) == 1
        assert result[0].id == sample_document.id
        assert result[0].content == sample_document.content
        assert builder.processing_stats["documents_processed"] == 1

    @patch('refinire_rag.processing.graph_builder.LLMPipeline')
    def test_process_with_llm_success(self, mock_llm_pipeline, config_with_temp_files, sample_document):
        """Test successful processing with LLM"""
        # Setup mock LLM response
        mock_result = Mock()
        mock_result.content = '''```json
{
    "has_new_relationships": true,
    "relationships_count": 3,
    "entities_count": 5,
    "extracted_relationships": [
        {
            "subject": "machine learning algorithms",
            "predicate": "process",
            "object": "data",
            "relationship_type": "functional"
        },
        {
            "subject": "API",
            "predicate": "connects_to",
            "object": "database",
            "relationship_type": "technical"
        },
        {
            "subject": "data preprocessing",
            "predicate": "improves",
            "object": "model accuracy",
            "relationship_type": "causal"
        }
    ],
    "entities": [
        {"name": "machine learning algorithms", "type": "concept"},
        {"name": "data", "type": "resource"},
        {"name": "API", "type": "component"},
        {"name": "database", "type": "component"},
        {"name": "model accuracy", "type": "metric"}
    ]
}
```'''
        
        mock_pipeline_instance = Mock()
        mock_pipeline_instance.run.return_value = mock_result
        mock_llm_pipeline.return_value = mock_pipeline_instance
        
        builder = GraphBuilder(config_with_temp_files)
        result = builder.process(sample_document)
        
        assert len(result) == 1
        assert result[0].id == sample_document.id
        assert result[0].content == sample_document.content
        assert builder.processing_stats["documents_processed"] == 1
        # Note: relationships_extracted may be 0 if mock data mode is used
        # or implementation doesn't update stats from LLM response
        
        # Check that graph file was created
        graph_path = Path(config_with_temp_files.graph_file_path)
        assert graph_path.exists()

    @patch('refinire_rag.processing.graph_builder.LLMPipeline')
    def test_process_with_llm_json_error(self, mock_llm_pipeline, config_with_temp_files, sample_document):
        """Test processing when LLM returns invalid JSON"""
        # Setup mock LLM response with invalid JSON
        mock_result = Mock()
        mock_result.content = "Invalid JSON response"
        
        mock_pipeline_instance = Mock()
        mock_pipeline_instance.run.return_value = mock_result
        mock_llm_pipeline.return_value = mock_pipeline_instance
        
        builder = GraphBuilder(config_with_temp_files)
        result = builder.process(sample_document)
        
        # Should still return document, but with error handling
        assert len(result) == 1
        assert result[0].id == sample_document.id
        assert result[0].content == sample_document.content
        assert builder.processing_stats["documents_processed"] == 1

    @patch('refinire_rag.processing.graph_builder.LLMPipeline')
    def test_process_with_llm_api_error(self, mock_llm_pipeline, config_with_temp_files, sample_document):
        """Test processing when LLM API call fails"""
        mock_pipeline_instance = Mock()
        mock_pipeline_instance.run.side_effect = Exception("API call failed")
        mock_llm_pipeline.return_value = mock_pipeline_instance
        
        builder = GraphBuilder(config_with_temp_files)
        result = builder.process(sample_document)
        
        # Should still return document, but with error handling
        assert len(result) == 1
        assert result[0].id == sample_document.id
        assert result[0].content == sample_document.content
        assert builder.processing_stats["documents_processed"] == 1

    def test_existing_graph_reading(self, config_with_temp_files):
        """Test reading existing graph file"""
        # Create existing graph file
        graph_path = Path(config_with_temp_files.graph_file_path)
        graph_path.parent.mkdir(parents=True, exist_ok=True)
        existing_content = "# Existing Graph\n\n## Entities\n- entity1\n\n## Relationships\n- rel1"
        graph_path.write_text(existing_content, encoding='utf-8')
        
        builder = GraphBuilder(config_with_temp_files)
        read_content = builder._read_existing_graph(config_with_temp_files)
        
        assert read_content == existing_content

    def test_nonexistent_graph_reading(self, config_with_temp_files):
        """Test reading when graph file doesn't exist"""
        builder = GraphBuilder(config_with_temp_files)
        content = builder._read_existing_graph(config_with_temp_files)
        
        # Should return empty graph template
        assert "# ドメイン知識グラフ" in content

    def test_dictionary_reading_from_config(self, config_with_temp_files, sample_dictionary_content):
        """Test reading dictionary from configured path"""
        # Create dictionary file
        dict_path = Path(config_with_temp_files.dictionary_file_path)
        dict_path.parent.mkdir(parents=True, exist_ok=True)
        dict_path.write_text(sample_dictionary_content, encoding='utf-8')
        
        builder = GraphBuilder(config_with_temp_files)
        sample_doc = Document(id="test", content="test", metadata={})
        content = builder._read_dictionary_content(sample_doc, config_with_temp_files)
        
        assert content == sample_dictionary_content

    def test_dictionary_reading_from_metadata(self, config_with_temp_files, sample_dictionary_content, temp_dir):
        """Test reading dictionary from document metadata"""
        # Create dictionary file in different location
        metadata_dict_path = temp_dir / "metadata_dict.md"
        metadata_dict_path.write_text(sample_dictionary_content, encoding='utf-8')
        
        # Document with dictionary metadata
        doc_with_metadata = Document(
            id="test",
            content="test",
            metadata={
                "dictionary_metadata": {
                    "dictionary_file_path": str(metadata_dict_path)
                }
            }
        )
        
        config_with_temp_files.auto_detect_dictionary_path = True
        builder = GraphBuilder(config_with_temp_files)
        content = builder._read_dictionary_content(doc_with_metadata, config_with_temp_files)
        
        assert content == sample_dictionary_content

    def test_dictionary_reading_fallback(self, config_with_temp_files):
        """Test dictionary reading when no files exist"""
        builder = GraphBuilder(config_with_temp_files)
        sample_doc = Document(id="test", content="test", metadata={})
        content = builder._read_dictionary_content(sample_doc, config_with_temp_files)
        
        assert content == ""

    def test_get_config_class(self):
        """Test get_config_class method"""
        assert GraphBuilder.get_config_class() == GraphBuilderConfig

    def test_processing_stats_initialization(self, config_with_temp_files):
        """Test that processing stats are properly initialized"""
        builder = GraphBuilder(config_with_temp_files)
        
        stats = builder.processing_stats
        assert "documents_processed" in stats
        assert "relationships_extracted" in stats
        assert "graph_updates" in stats
        assert "llm_api_calls" in stats
        assert "duplicate_relationships_avoided" in stats
        assert stats["documents_processed"] == 0

    def test_backup_functionality(self, config_with_temp_files, temp_dir):
        """Test graph backup functionality"""
        config_with_temp_files.backup_graph = True
        
        # Create existing graph
        graph_path = Path(config_with_temp_files.graph_file_path)
        graph_path.parent.mkdir(parents=True, exist_ok=True)
        graph_path.write_text("Original graph content", encoding='utf-8')
        
        builder = GraphBuilder(config_with_temp_files)
        builder._backup_graph_file(config_with_temp_files)
        
        # Check backup was created - backup files use format: .backup_timestamp.md
        backup_files = list(graph_path.parent.glob(f"{graph_path.stem}.backup_*.md"))
        assert len(backup_files) > 0

    def test_relationship_extraction_prompt_creation(self, config_with_temp_files, sample_document, sample_dictionary_content):
        """Test creation of relationship extraction prompt"""
        builder = GraphBuilder(config_with_temp_files)
        
        existing_graph = "# Existing Graph Content"
        prompt = builder._create_relationship_extraction_prompt(
            sample_document, existing_graph, sample_dictionary_content, config_with_temp_files
        )
        
        # Check that prompt contains key elements (in Japanese)
        assert "重要な関係性" in prompt  # "important relationships" in Japanese
        assert sample_document.content in prompt
        assert existing_graph in prompt
        assert sample_dictionary_content in prompt

    def test_skip_processing_config(self, config_with_temp_files, sample_document):
        """Test skip processing configuration"""
        config_with_temp_files.skip_if_no_new_relationships = True
        
        builder = GraphBuilder(config_with_temp_files)
        builder._llm_pipeline = None  # Force mock mode
        
        result = builder.process(sample_document)
        
        assert len(result) == 1
        assert result[0].id == sample_document.id
        assert result[0].content == sample_document.content


if __name__ == "__main__":
    pytest.main([__file__])


"""
FileInfo model for tracking file state
ファイル状態追跡用のFileInfoモデル
"""

import hashlib
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Dict, Any
from refinire_rag.models.document import Document

@dataclass
class FileInfo:
    """
    Information about a file for incremental loading
    インクリメンタルローディング用のファイル情報
    
    This class holds the essential information needed to track changes
    in files between scans.
    このクラスは、スキャン間でファイルの変更を追跡するために
    必要な基本情報を保持します。
    """
    path: str
    size: int  
    modified_at: datetime
    hash_md5: str
    file_type: str
    
    @classmethod
    def from_file(cls, file_path: Path) -> 'FileInfo':
        """
        Create FileInfo from a file on disk
        ディスク上のファイルからFileInfoを作成
        
        Args:
            file_path: Path to the file
            file_path: ファイルのパス
            
        Returns:
            FileInfo object with file information
            ファイル情報を含むFileInfoオブジェクト
        """
        stat = file_path.stat()
        
        # Calculate MD5 hash of file content
        # ファイル内容のMD5ハッシュを計算
        hash_md5 = cls._calculate_file_hash(file_path)
        
        # Determine file type from extension
        # 拡張子からファイルタイプを決定
        file_type = cls._determine_file_type(file_path)
        
        return cls(
            path=str(file_path),
            size=stat.st_size,
            modified_at=datetime.fromtimestamp(stat.st_mtime),
            hash_md5=hash_md5,
            file_type=file_type
        )
    
    @classmethod
    def from_document_metadata(cls, metadata: Dict[str, Any]) -> 'FileInfo':
        """
        Create FileInfo from document metadata
        文書メタデータからFileInfoを作成
        
        Args:
            metadata: Document metadata dictionary
            metadata: 文書メタデータ辞書
            
        Returns:
            FileInfo object reconstructed from metadata
            メタデータから再構築されたFileInfoオブジェクト
        """
        return cls(
            path=metadata.get('file_path', ''),
            size=metadata.get('file_size', 0),
            modified_at=datetime.fromisoformat(metadata.get('file_modified_at', datetime.now().isoformat())),
            hash_md5=metadata.get('file_hash', ''),
            file_type=metadata.get('file_type', 'unknown')
        )
    
    @staticmethod
    def _calculate_file_hash(file_path: Path) -> str:
        """
        Calculate MD5 hash of file content
        ファイル内容のMD5ハッシュを計算
        
        Args:
            file_path: Path to the file
            file_path: ファイルのパス
            
        Returns:
            MD5 hash as hexadecimal string
            16進文字列としてのMD5ハッシュ
        """
        hash_md5 = hashlib.md5()
        try:
            with open(file_path, "rb") as f:
                # Read file in chunks to handle large files efficiently
                # 大きなファイルを効率的に処理するためにチャンクで読み込み
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
        except (OSError, IOError):
            # If file cannot be read, return empty hash
            # ファイルが読み込めない場合は空のハッシュを返す
            return ""
        
        return hash_md5.hexdigest()
    
    @staticmethod
    def _determine_file_type(file_path: Path) -> str:
        """
        Determine file type from extension
        拡張子からファイルタイプを決定
        
        Args:
            file_path: Path to the file
            file_path: ファイルのパス
            
        Returns:
            File type string
            ファイルタイプ文字列
        """
        extension = file_path.suffix.lower()
        
        # Map common extensions to file types
        # 一般的な拡張子をファイルタイプにマッピング
        extension_map = {
            '.txt': 'text',
            '.md': 'markdown',
            '.csv': 'csv',
            '.json': 'json',
            '.html': 'html',
            '.htm': 'html',
            '.pdf': 'pdf',
            '.doc': 'document',
            '.docx': 'document',
            '.py': 'code',
            '.js': 'code',
            '.ts': 'code',
            '.java': 'code',
            '.cpp': 'code',
            '.c': 'code',
        }
        
        return extension_map.get(extension, 'unknown')
    
    def __eq__(self, other) -> bool:
        """
        Check equality based on hash, size, and modification time
        ハッシュ、サイズ、更新時間に基づく等価性をチェック
        """
        if not isinstance(other, FileInfo):
            return False
        
        return (
            self.hash_md5 == other.hash_md5 and
            self.size == other.size and
            self.modified_at == other.modified_at
        )
    
    def __hash__(self) -> int:
        """
        Generate hash for use in sets and dictionaries
        セットや辞書で使用するためのハッシュを生成
        """
        return hash((self.path, self.hash_md5, self.size, self.modified_at))


"""
Size-based document splitting processor
サイズベースの文書分割プロセッサー
"""

import logging
from typing import Iterator, Iterable, Optional, Any
from refinire_rag.splitter.splitter import Splitter
from refinire_rag.models.document import Document
from refinire_rag.utils import generate_chunk_id

logger = logging.getLogger(__name__)

class SizeSplitter(Splitter):
    """Processor that splits documents into chunks based on size
    サイズベースで文書を分割するプロセッサー"""
    
    def __init__(
        self,
        chunk_size: int = 1024,  # 1KB default
        overlap_size: int = 0    # No overlap by default
    ):
        """Initialize size splitter
        サイズ分割プロセッサーを初期化
        
        Args:
            chunk_size: Maximum size of each chunk in bytes
            チャンクサイズ: 各チャンクの最大サイズ（バイト）
            overlap_size: Size of overlap between chunks in bytes
            オーバーラップサイズ: チャンク間のオーバーラップサイズ（バイト）
        """
        super().__init__({
            'chunk_size': chunk_size,
            'overlap_size': overlap_size
        })
    
    def process(self, documents: Iterable[Document], config: Optional[Any] = None) -> Iterator[Document]:
        """Split documents into chunks based on size
        文書をサイズベースで分割
        
        Args:
            documents: Input documents to process
            config: Optional configuration for splitting
            
        Yields:
            Split documents
        """
        chunk_config = config or self.config
        chunk_size = chunk_config.get('chunk_size', 1024)
        overlap_size = chunk_config.get('overlap_size', 0)
        
        # Prevent infinite loop if overlap_size >= chunk_size
        # overlap_sizeがchunk_size以上の場合は無限ループを防ぐ
        if overlap_size >= chunk_size:
            logger.warning(
                "overlap_size (%d) >= chunk_size (%d). Setting overlap_size = chunk_size - 1.",
                overlap_size, chunk_size
            )
            overlap_size = chunk_size - 1 if chunk_size > 1 else 0
        
        for doc in documents:
            content = doc.content
            content_size = len(content.encode('utf-8'))
            
            if content_size <= chunk_size:
                # If document is smaller than chunk size, yield as is
                yield doc
                continue
            
            # Split content into chunks
            start = 0
            chunk_index = 0
            while start < len(content):
                # Calculate chunk end position
                end = start + chunk_size
                if end > len(content):
                    end = len(content)
                
                # Extract chunk
                chunk_content = content[start:end]
                
                # Create new document for chunk
                chunk_doc = Document(
                    id=generate_chunk_id(),
                    content=chunk_content,
                    metadata={
                        **doc.metadata,
                        'chunk_index': chunk_index,
                        'total_chunks': (len(content) + chunk_size - 1) // chunk_size,
                        'chunk_start': start,
                        'chunk_end': end,
                        'origin_id': doc.id,
                        'original_document_id': doc.id
                    }
                )
                
                yield chunk_doc
                
                # Move to next chunk, considering overlap
                next_start = end - overlap_size
                if next_start <= start:
                    next_start = start + 1
                start = next_start
                chunk_index += 1 


"""
Pickle-based Vector Store Implementation

Persistent vector storage using pickle serialization.
Good for development and small to medium datasets that need persistence.
"""

import pickle
import logging
from pathlib import Path
from typing import List, Dict, Any, Optional, Type
import numpy as np

from .in_memory_vector_store import InMemoryVectorStore
from .vector_store import VectorEntry, VectorSearchResult, VectorStoreStats
from ..exceptions import StorageError

logger = logging.getLogger(__name__)


class PickleVectorStore(InMemoryVectorStore):
    """Persistent vector store using pickle serialization
    
    Extends InMemoryVectorStore with automatic persistence to disk.
    Data is kept in memory for fast access and periodically saved to disk.
    """
    
    def __init__(
        self, 
        file_path: str = "./data/vectors.pkl",
        similarity_metric: str = "cosine",
        auto_save: bool = True,
        save_interval: int = 10,
        config: Optional[Dict] = None
    ):
        """Initialize pickle-based vector store
        
        Args:
            file_path: Path to pickle file for persistence
            similarity_metric: Similarity metric to use
            auto_save: Whether to automatically save changes
            save_interval: Number of operations before auto-save (if auto_save=True)
            config: Optional configuration for DocumentProcessor
        """
        pickle_config = config or {
            "file_path": file_path,
            "similarity_metric": similarity_metric,
            "auto_save": auto_save,
            "save_interval": save_interval
        }
        super().__init__(similarity_metric=similarity_metric, config=pickle_config)
        
        self.file_path = Path(file_path)
        self.file_path.parent.mkdir(parents=True, exist_ok=True)
        
        self.auto_save = auto_save
        self.save_interval = save_interval
        self._operations_since_save = 0
        
        # Load existing data
        self._load_from_disk()
        
        logger.info(f"Initialized PickleVectorStore at {self.file_path}")
    
    def add_vector(self, entry: VectorEntry) -> str:
        """Add a vector entry and optionally save to disk"""
        
        result = super().add_vector(entry)
        self._maybe_save()
        return result
    
    def add_vectors(self, entries: List[VectorEntry]) -> List[str]:
        """Add multiple vector entries and optionally save to disk"""
        
        result = super().add_vectors(entries)
        self._maybe_save()
        return result
    
    def update_vector(self, entry: VectorEntry) -> bool:
        """Update vector entry and optionally save to disk"""
        
        result = super().update_vector(entry)
        if result:
            self._maybe_save()
        return result
    
    def delete_vector(self, document_id: str) -> bool:
        """Delete vector entry and optionally save to disk"""
        
        result = super().delete_vector(document_id)
        if result:
            self._maybe_save()
        return result
    
    def clear(self) -> bool:
        """Clear all vectors and save to disk"""
        
        result = super().clear()
        if result:
            self.save_to_disk()
        return result
    
    def save_to_disk(self) -> bool:
        """Explicitly save current state to disk"""
        
        try:
            # Prepare data for serialization
            data = {
                'vectors': self._vectors,
                'similarity_metric': self.similarity_metric,
                'metadata': {
                    'total_vectors': len(self._vectors),
                    'vector_dimension': self.get_vector_dimension(),
                    'save_timestamp': np.datetime64('now')
                }
            }
            
            # Write to temporary file first, then rename (atomic operation)
            temp_path = self.file_path.with_suffix('.tmp')
            with open(temp_path, 'wb') as f:
                pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
            
            # Atomic rename
            temp_path.replace(self.file_path)
            
            self._operations_since_save = 0
            logger.info(f"Saved {len(self._vectors)} vectors to {self.file_path}")
            return True
            
        except Exception as e:
            logger.error(f"Failed to save vectors to disk: {e}")
            return False
    
    def load_from_disk(self) -> bool:
        """Explicitly load state from disk"""
        return self._load_from_disk()
    
    def _load_from_disk(self) -> bool:
        """Load vectors from pickle file"""
        
        try:
            if not self.file_path.exists():
                logger.info(f"No existing vector file found at {self.file_path}")
                return True
            
            with open(self.file_path, 'rb') as f:
                data = pickle.load(f)
            
            # Restore vectors
            self._vectors = data.get('vectors', {})
            
            # Restore similarity metric if compatible
            saved_metric = data.get('similarity_metric', 'cosine')
            if saved_metric != self.similarity_metric:
                logger.warning(f"Similarity metric mismatch: loaded '{saved_metric}', expected '{self.similarity_metric}'")
            
            # Force matrix rebuild
            self._needs_rebuild = True
            
            # Log metadata if available
            metadata = data.get('metadata', {})
            if metadata:
                logger.info(f"Loaded {metadata.get('total_vectors', 0)} vectors from {self.file_path}")
                if 'save_timestamp' in metadata:
                    logger.debug(f"Data saved at: {metadata['save_timestamp']}")
            else:
                logger.info(f"Loaded {len(self._vectors)} vectors from {self.file_path}")
            
            return True
            
        except Exception as e:
            logger.error(f"Failed to load vectors from disk: {e}")
            # Continue with empty store
            self._vectors = {}
            return False
    
    def _maybe_save(self):
        """Save to disk if auto_save is enabled and save_interval is reached"""
        
        if not self.auto_save:
            return
        
        self._operations_since_save += 1
        
        if self._operations_since_save >= self.save_interval:
            self.save_to_disk()
    
    def get_stats(self) -> VectorStoreStats:
        """Get vector store statistics including file size"""
        
        stats = super().get_stats()
        
        # Add file size information
        if self.file_path.exists():
            file_size = self.file_path.stat().st_size
            stats.storage_size_bytes = file_size
        
        # Update index type
        stats.index_type = "exact_pickle"
        
        return stats
    
    def backup_to_file(self, backup_path: str) -> bool:
        """Create a backup of the vector store"""
        
        try:
            backup_file = Path(backup_path)
            backup_file.parent.mkdir(parents=True, exist_ok=True)
            
            # Save current state first
            self.save_to_disk()
            
            # Copy to backup location
            import shutil
            shutil.copy2(self.file_path, backup_file)
            
            logger.info(f"Created backup at {backup_path}")
            return True
            
        except Exception as e:
            logger.error(f"Failed to create backup: {e}")
            return False
    
    def restore_from_backup(self, backup_path: str) -> bool:
        """Restore vector store from a backup"""
        
        try:
            backup_file = Path(backup_path)
            if not backup_file.exists():
                logger.error(f"Backup file {backup_path} does not exist")
                return False
            
            # Copy backup to main file
            import shutil
            shutil.copy2(backup_file, self.file_path)
            
            # Reload from disk
            return self._load_from_disk()
            
        except Exception as e:
            logger.error(f"Failed to restore from backup: {e}")
            return False
    
    def optimize_storage(self) -> bool:
        """Optimize storage (defragment, remove unused space)"""
        
        try:
            # For pickle store, this is simply saving to disk
            # which rewrites the entire file
            return self.save_to_disk()
            
        except Exception as e:
            logger.error(f"Failed to optimize storage: {e}")
            return False
    
    def get_file_info(self) -> Dict[str, Any]:
        """Get information about the storage file"""
        
        try:
            if not self.file_path.exists():
                return {
                    "exists": False,
                    "path": str(self.file_path)
                }
            
            stat = self.file_path.stat()
            
            return {
                "exists": True,
                "path": str(self.file_path),
                "size_bytes": stat.st_size,
                "modified_time": stat.st_mtime,
                "operations_since_save": self._operations_since_save,
                "auto_save": self.auto_save,
                "save_interval": self.save_interval
            }
            
        except Exception as e:
            logger.error(f"Failed to get file info: {e}")
            return {"error": str(e)}
    
    def close(self):
        """Save and close the vector store"""
        if self.auto_save and self._operations_since_save > 0:
            self.save_to_disk()
        logger.info("PickleVectorStore closed")
    
    def __del__(self):
        """Cleanup on deletion"""
        try:
            self.close()
        except:
            pass


"""
DocumentStore Loader for loading documents from existing DocumentStore
既存のDocumentStoreから文書をロードするDocumentStoreLoader
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Any, Iterator, Iterable
from datetime import datetime

from refinire_rag.loader.loader import Loader
from refinire_rag.storage.document_store import DocumentStore
from refinire_rag.models.document import Document
from refinire_rag.metadata.metadata import Metadata
from refinire_rag.exceptions import (
    DocumentStoreError, LoaderError, ValidationError, 
    ConfigurationError, wrap_exception
)


class LoadStrategy(Enum):
    """
    Document loading strategies
    文書ロード戦略
    """
    FULL = "full"           # Load all documents / すべての文書をロード
    FILTERED = "filtered"   # Load with metadata/content filters / メタデータ・内容フィルターでロード
    INCREMENTAL = "incremental"  # Load based on timestamps / タイムスタンプベースでロード
    ID_LIST = "id_list"     # Load specific document IDs / 特定の文書IDをロード
    PAGINATED = "paginated" # Load in batches / バッチでロード


@dataclass
class DocumentLoadConfig:
    """
    Configuration for document loading from DocumentStore
    DocumentStoreからの文書ロード設定
    """
    # Loading strategy
    # ロード戦略
    strategy: LoadStrategy = LoadStrategy.FULL
    
    # Filtering options
    # フィルタリングオプション
    metadata_filters: Optional[Dict[str, Any]] = None
    content_query: Optional[str] = None
    document_ids: Optional[List[str]] = None
    
    # Date-based filtering
    # 日付ベースフィルタリング
    modified_after: Optional[datetime] = None
    modified_before: Optional[datetime] = None
    created_after: Optional[datetime] = None
    created_before: Optional[datetime] = None
    
    # Pagination
    # ページング
    batch_size: int = 100
    max_documents: Optional[int] = None
    
    # Sorting
    # ソート
    sort_by: str = "created_at"
    sort_order: str = "desc"
    
    # Processing options
    # 処理オプション
    include_deleted: bool = False
    validate_documents: bool = True
    
    def validate(self) -> None:
        """
        Validate configuration settings
        設定の妥当性を検証
        """
        if self.batch_size <= 0:
            raise ValidationError("batch_size must be positive")
        
        if self.max_documents is not None and self.max_documents <= 0:
            raise ValidationError("max_documents must be positive")
        
        if self.strategy == LoadStrategy.ID_LIST and not self.document_ids:
            raise ConfigurationError("document_ids required for ID_LIST strategy")
        
        if self.modified_after and self.modified_before:
            if self.modified_after >= self.modified_before:
                raise ValidationError("modified_after must be before modified_before")
        
        if self.created_after and self.created_before:
            if self.created_after >= self.created_before:
                raise ValidationError("created_after must be before created_before")
        
        if self.sort_order not in ["asc", "desc"]:
            raise ValidationError("sort_order must be 'asc' or 'desc'")


@dataclass
class LoadResult:
    """
    Result of document loading operation
    文書ロード操作の結果
    """
    loaded_count: int = 0
    skipped_count: int = 0
    error_count: int = 0
    errors: List[str] = field(default_factory=list)
    total_processed: int = 0
    
    def add_error(self, error_message: str):
        """
        Add error message and increment error count
        エラーメッセージを追加してエラー数を増加
        """
        self.errors.append(error_message)
        self.error_count += 1
    
    @property
    def success_rate(self) -> float:
        """
        Calculate success rate
        成功率を計算
        """
        if self.total_processed == 0:
            return 1.0
        return (self.loaded_count + self.skipped_count) / self.total_processed
    
    def get_summary(self) -> Dict[str, Any]:
        """
        Get summary of loading results
        ロード結果のサマリーを取得
        """
        return {
            "loaded": self.loaded_count,
            "skipped": self.skipped_count,
            "errors": self.error_count,
            "total_processed": self.total_processed,
            "success_rate": self.success_rate,
            "error_messages": self.errors[-5:] if self.errors else []  # Last 5 errors
        }


class DocumentStoreLoader(Loader):
    """
    Loader for documents from DocumentStore
    DocumentStoreからの文書ローダー
    
    This loader can load documents from an existing DocumentStore using various
    strategies including full load, filtered load, incremental load, and more.
    このローダーは既存のDocumentStoreからフルロード、フィルタードロード、
    インクリメンタルロードなどの様々な戦略を使用して文書をロードできます。
    """
    
    def __init__(self, 
                 document_store: DocumentStore,
                 load_config: Optional[DocumentLoadConfig] = None,
                 metadata_processors: Optional[List[Metadata]] = None):
        """
        Initialize DocumentStore loader
        DocumentStoreローダーを初期化
        
        Args:
            document_store: DocumentStore instance to load from
            load_config: Loading configuration
            metadata_processors: Optional metadata processors
            document_store: ロード元のDocumentStoreインスタンス
            load_config: ロード設定
            metadata_processors: オプションのメタデータプロセッサー
        """
        super().__init__(metadata_processors)
        
        if document_store is None:
            raise ConfigurationError("document_store cannot be None")
        
        self.document_store = document_store
        self.load_config = load_config or DocumentLoadConfig()
        self.metadata_processors = metadata_processors or []
        
        # Validate configuration
        # 設定を検証
        try:
            self.load_config.validate()
        except (ValidationError, ConfigurationError) as e:
            raise ConfigurationError(f"Invalid load configuration: {e}")
    
    def process(self, documents: Iterable[Document], config: Optional[Any] = None) -> Iterator[Document]:
        """
        DocumentProcessor interface implementation
        DocumentProcessorインターフェース実装
        
        This method loads documents from the store and yields them,
        ignoring the input documents parameter.
        このメソッドはストアから文書をロードしてyieldし、
        入力documentsパラメータは無視します。
        
        Args:
            documents: Input documents (ignored)
            config: Optional configuration (ignored)
            documents: 入力文書（無視される）
            config: オプション設定（無視される）
            
        Yields:
            Documents loaded from the store
            ストアからロードされた文書
        """
        try:
            for document in self._load_documents():
                yield document
        except Exception as e:
            if isinstance(e, (DocumentStoreError, LoaderError)):
                raise
            else:
                raise wrap_exception(e, "Error in document processing")
    
    def load_all(self) -> LoadResult:
        """
        Load all documents matching configuration
        設定にマッチするすべての文書をロード
        
        Returns:
            LoadResult with loading statistics
            ロード統計を含むLoadResult
        """
        result = LoadResult()
        
        try:
            for document in self._load_documents():
                try:
                    if self._validate_document(document):
                        result.loaded_count += 1
                    else:
                        result.skipped_count += 1
                except Exception as e:
                    result.add_error(f"Error processing document {document.id}: {str(e)}")
                finally:
                    result.total_processed += 1
                    
        except Exception as e:
            error_msg = f"Failed to load documents: {str(e)}"
            result.add_error(error_msg)
            if isinstance(e, (DocumentStoreError, LoaderError)):
                raise
            else:
                raise wrap_exception(e, "Document loading failed")
        
        return result
    
    def _load_documents(self) -> Iterator[Document]:
        """
        Load documents based on configuration strategy
        設定戦略に基づいて文書をロード
        
        Yields:
            Documents from the store
            ストアからの文書
        """
        try:
            if self.load_config.strategy == LoadStrategy.FULL:
                yield from self._load_all_documents()
            elif self.load_config.strategy == LoadStrategy.FILTERED:
                yield from self._load_filtered_documents()
            elif self.load_config.strategy == LoadStrategy.INCREMENTAL:
                yield from self._load_incremental_documents()
            elif self.load_config.strategy == LoadStrategy.ID_LIST:
                yield from self._load_by_ids()
            elif self.load_config.strategy == LoadStrategy.PAGINATED:
                yield from self._load_paginated_documents()
            else:
                raise LoaderError(f"Unsupported load strategy: {self.load_config.strategy}")
                
        except DocumentStoreError:
            # Re-raise DocumentStore errors as-is
            # DocumentStoreエラーはそのまま再発生
            raise
        except Exception as e:
            raise wrap_exception(e, f"Error loading documents with strategy {self.load_config.strategy}")
    
    def _load_all_documents(self) -> Iterator[Document]:
        """
        Load all documents from store
        ストアからすべての文書をロード
        """
        try:
            offset = 0
            total_loaded = 0
            
            while True:
                documents = self.document_store.list_documents(
                    limit=self.load_config.batch_size,
                    offset=offset,
                    sort_by=self.load_config.sort_by,
                    sort_order=self.load_config.sort_order
                )
                
                if not documents:
                    break
                
                for doc in documents:
                    yield doc
                    total_loaded += 1
                    
                    # Check max_documents limit
                    # max_documents制限をチェック
                    if (self.load_config.max_documents and 
                        total_loaded >= self.load_config.max_documents):
                        return
                    
                offset += len(documents)
                
        except Exception as e:
            raise wrap_exception(e, "Error loading all documents")
    
    def _load_filtered_documents(self) -> Iterator[Document]:
        """
        Load documents with metadata/content filters
        メタデータ・内容フィルターで文書をロード
        """
        try:
            if self.load_config.metadata_filters or self._has_date_filters():
                # Build complete filter including date filters
                # 日付フィルターを含む完全なフィルターを構築
                filters = self._build_metadata_filters()
                search_results = self.document_store.search_by_metadata(
                    filters=filters,
                    limit=self.load_config.max_documents or 1000000
                )
                for result in search_results:
                    yield result.document
                    
            elif self.load_config.content_query:
                search_results = self.document_store.search_by_content(
                    query=self.load_config.content_query,
                    limit=self.load_config.max_documents or 1000000
                )
                for result in search_results:
                    yield result.document
            else:
                # No specific filters, load all
                # 特定のフィルターなし、すべてをロード
                yield from self._load_all_documents()
                
        except Exception as e:
            raise wrap_exception(e, "Error loading filtered documents")
    
    def _load_incremental_documents(self) -> Iterator[Document]:
        """
        Load documents based on modification timestamps
        更新タイムスタンプに基づいて文書をロード
        """
        try:
            filters = {}
            
            if self.load_config.modified_after:
                filters['modified_at'] = {'$gte': self.load_config.modified_after.isoformat()}
            
            if self.load_config.modified_before:
                if 'modified_at' in filters:
                    filters['modified_at']['$lte'] = self.load_config.modified_before.isoformat()
                else:
                    filters['modified_at'] = {'$lte': self.load_config.modified_before.isoformat()}
            
            if not filters:
                raise LoaderError("No timestamp filters specified for incremental loading")
            
            search_results = self.document_store.search_by_metadata(
                filters=filters,
                limit=self.load_config.max_documents or 1000000
            )
            
            for result in search_results:
                yield result.document
                
        except LoaderError:
            raise
        except Exception as e:
            raise wrap_exception(e, "Error loading incremental documents")
    
    def _load_by_ids(self) -> Iterator[Document]:
        """
        Load specific documents by IDs
        特定のIDで文書をロード
        """
        if not self.load_config.document_ids:
            raise LoaderError("No document IDs specified for ID list loading")
        
        try:
            for doc_id in self.load_config.document_ids:
                document = self.document_store.get_document(doc_id)
                if document:
                    yield document
                elif self.load_config.validate_documents:
                    raise LoaderError(f"Document not found: {doc_id}")
                    
        except LoaderError:
            raise
        except Exception as e:
            raise wrap_exception(e, "Error loading documents by IDs")
    
    def _load_paginated_documents(self) -> Iterator[Document]:
        """
        Load documents in paginated fashion
        ページング方式で文書をロード
        """
        try:
            offset = 0
            total_loaded = 0
            
            while True:
                documents = self.document_store.list_documents(
                    limit=self.load_config.batch_size,
                    offset=offset,
                    sort_by=self.load_config.sort_by,
                    sort_order=self.load_config.sort_order
                )
                
                if not documents:
                    break
                
                for doc in documents:
                    yield doc
                    total_loaded += 1
                    
                    if (self.load_config.max_documents and 
                        total_loaded >= self.load_config.max_documents):
                        return
                
                offset += len(documents)
                
        except Exception as e:
            raise wrap_exception(e, "Error loading paginated documents")
    
    def _has_date_filters(self) -> bool:
        """
        Check if any date filters are configured
        日付フィルターが設定されているかチェック
        """
        return any([
            self.load_config.modified_after,
            self.load_config.modified_before,
            self.load_config.created_after,
            self.load_config.created_before
        ])
    
    def _build_metadata_filters(self) -> Dict[str, Any]:
        """
        Build complete metadata filters from configuration
        設定から完全なメタデータフィルターを構築
        """
        filters = {}
        
        if self.load_config.metadata_filters:
            filters.update(self.load_config.metadata_filters)
        
        # Add date-based filters
        # 日付ベースフィルターを追加
        if self.load_config.modified_after or self.load_config.modified_before:
            date_filter = {}
            if self.load_config.modified_after:
                date_filter['$gte'] = self.load_config.modified_after.isoformat()
            if self.load_config.modified_before:
                date_filter['$lte'] = self.load_config.modified_before.isoformat()
            filters['modified_at'] = date_filter
        
        if self.load_config.created_after or self.load_config.created_before:
            date_filter = {}
            if self.load_config.created_after:
                date_filter['$gte'] = self.load_config.created_after.isoformat()
            if self.load_config.created_before:
                date_filter['$lte'] = self.load_config.created_before.isoformat()
            filters['created_at'] = date_filter
        
        return filters
    
    def _validate_document(self, document: Document) -> bool:
        """
        Validate document before yielding
        文書をyieldする前に検証
        
        Args:
            document: Document to validate
            document: 検証する文書
            
        Returns:
            True if document is valid, False otherwise
            文書が有効な場合True、そうでなければFalse
        """
        if not self.load_config.validate_documents:
            return True
        
        try:
            # Basic validation
            # 基本的な検証
            if not document.id:
                raise ValidationError("Document missing ID")
            
            if not document.content and not document.metadata:
                raise ValidationError("Document has no content or metadata")
            
            return True
            
        except ValidationError as e:
            if self.load_config.validate_documents:
                raise LoaderError(f"Document validation failed: {e}")
            return False
    
    def count_matching_documents(self) -> int:
        """
        Count documents that would be loaded with current configuration
        現在の設定でロードされる文書数をカウント
        
        Returns:
            Number of matching documents, or -1 if unknown
            マッチする文書数、不明な場合は-1
        """
        try:
            if self.load_config.strategy == LoadStrategy.FULL:
                return self.document_store.count_documents()
            elif self.load_config.strategy == LoadStrategy.FILTERED:
                if self.load_config.metadata_filters or self._has_date_filters():
                    filters = self._build_metadata_filters()
                    return self.document_store.count_documents(filters)
                else:
                    return self.document_store.count_documents()
            elif self.load_config.strategy == LoadStrategy.ID_LIST:
                return len(self.load_config.document_ids or [])
            else:
                # For other strategies, we'd need to actually query
                # 他の戦略では実際にクエリが必要
                return -1  # Unknown
                
        except Exception as e:
            raise wrap_exception(e, "Error counting matching documents")
    
    def get_load_summary(self) -> Dict[str, Any]:
        """
        Get summary of loader configuration and capabilities
        ローダー設定と機能のサマリーを取得
        
        Returns:
            Dictionary with loader information
            ローダー情報を含む辞書
        """
        try:
            return {
                "strategy": self.load_config.strategy.value,
                "batch_size": self.load_config.batch_size,
                "max_documents": self.load_config.max_documents,
                "has_metadata_filters": bool(self.load_config.metadata_filters),
                "has_content_query": bool(self.load_config.content_query),
                "has_date_filters": self._has_date_filters(),
                "validate_documents": self.load_config.validate_documents,
                "estimated_count": self.count_matching_documents()
            }
        except Exception as e:
            return {"error": str(e)}
    
    def __str__(self) -> str:
        """
        String representation of the loader
        ローダーの文字列表現
        """
        return f"DocumentStoreLoader(strategy={self.load_config.strategy.value})"
    
    def __repr__(self) -> str:
        """
        Developer representation of the loader
        ローダーの開発者向け表現
        """
        return (f"DocumentStoreLoader(strategy={self.load_config.strategy.value}, "
                f"batch_size={self.load_config.batch_size}, "
                f"validate={self.load_config.validate_documents})")


"""
MarkdownTextSplitter - Markdown document splitter

A DocumentProcessor that splits markdown documents into chunks while preserving
markdown structure and formatting.
"""

import logging
from typing import List, Optional, Type, Any, Iterator, Dict
import re

from refinire_rag.models.document import Document
from refinire_rag.splitter.splitter import Splitter
from refinire_rag.utils import generate_chunk_id

logger = logging.getLogger(__name__)

class MarkdownTextSplitter(Splitter):
    """Markdownテキストを構造を保持しながら分割するスプリッター
    
    Markdownの構造（見出し、リスト、コードブロックなど）を保持しながら、
    テキストを適切なサイズのチャンクに分割します。
    """
    
    def __init__(self, chunk_size: int = 1000, overlap_size: int = 200):
        """
        Initialize the MarkdownTextSplitter
        MarkdownTextSplitterを初期化
        Args:
            chunk_size: Maximum size of each chunk (in characters)
            chunk_size: 各チャンクの最大サイズ（文字数）
            overlap_size: Overlap size between chunks (in characters)
            overlap_size: チャンク間のオーバーラップ（文字数）
        """
        super().__init__(chunk_size, overlap_size)
        self.processing_stats = {
            "headers_preserved": 0,
            "lists_preserved": 0,
            "code_blocks_preserved": 0,
            "tables_preserved": 0
        }
        
    def process(self, documents: List[Document]) -> Iterator[Document]:
        """
        Process documents and split them into chunks
        ドキュメントを処理し、チャンクに分割
        Args:
            documents: List of documents to process
            documents: 処理対象のドキュメントリスト
        Returns:
            Iterator of processed documents
            処理済みドキュメントのイテレータ
        """
        for doc in documents:
            if not doc.content.strip():
                yield doc
                continue
            
            chunks = self._split_text(doc.content)
            total_chunks = len(chunks)
            
            for i, chunk in enumerate(chunks):
                metadata = doc.metadata.copy() if doc.metadata else {}
                metadata.update({
                    "original_document_id": doc.id,
                    "chunk_index": i,
                    "total_chunks": total_chunks,
                    "section_index": i // self.chunk_size,
                    "total_sections": (total_chunks + self.chunk_size - 1) // self.chunk_size
                })
                yield Document(
                    content=chunk,
                    id=generate_chunk_id(),
                    metadata={
                        **metadata,
                        'origin_id': doc.id,
                        'original_document_id': doc.id
                    }
                )
                
    def _split_by_headers(self, content: str) -> List[str]:
        """
        Split text by markdown headers (#~######)
        Markdownの見出しでテキストを分割
        Args:
            content: Text to split
            content: 分割対象のテキスト
        Returns:
            List of sections split by headers
            見出しで分割されたテキストのリスト
        """
        header_pattern = r'^#{1,6}\s+.+$'
        lines = content.split('\n')
        sections = []
        current_section = []
        for line in lines:
            if re.match(header_pattern, line):
                if current_section:
                    sections.append('\n'.join(current_section))
                    current_section = []
            current_section.append(line)
        if current_section:
            sections.append('\n'.join(current_section))
        return sections
        
    def _split_text(self, text: str) -> List[str]:
        """
        Split text into chunks, always grouping header and its following text (including blank lines) into a single chunk.
        ヘッダーとその直後のテキスト（空行も含む）を常に1チャンクにまとめる
        Args:
            text: Text to split
            text: 分割対象のテキスト
        Returns:
            List of text chunks
            チャンク化されたテキストのリスト
        """
        lines = text.split('\n')
        chunks = []
        current_chunk = []
        i = 0
        while i < len(lines):
            line = lines[i]
            if line.startswith('#'):
                # 既存のチャンクがあれば追加
                if current_chunk:
                    chunks.append('\n'.join(current_chunk))
                    current_chunk = []
                # 新しいチャンクを開始
                current_chunk.append(line)
                i += 1
                # 次のヘッダーまたはファイル末尾までをこのチャンクに追加
                while i < len(lines) and not lines[i].startswith('#'):
                    current_chunk.append(lines[i])
                    i += 1
                # チャンクを追加
                chunks.append('\n'.join(current_chunk))
                current_chunk = []
            else:
                # ヘッダー以外の行（ファイル先頭やヘッダーなしの場合）
                current_chunk.append(line)
                i += 1
        if current_chunk:
            chunks.append('\n'.join(current_chunk))
        # 空チャンクは除外
        return [c for c in chunks if c.strip()]

    def _split_section(self, text: str) -> List[str]:
        """
        Split a section of text into chunks with overlap
        セクション内のテキストをオーバーラップ付きでチャンクに分割
        Args:
            text: Text to split
            text: 分割対象のテキスト
        Returns:
            List of text chunks
            チャンク化されたテキストのリスト
        """
        if not text.strip():
            return [text]
        
        # 文単位で分割
        sentences = []
        current_sentence = []
        for line in text.split('\n'):
            if line.strip():
                current_sentence.append(line)
            else:
                if current_sentence:
                    sentences.append('\n'.join(current_sentence))
                    current_sentence = []
        if current_sentence:
            sentences.append('\n'.join(current_sentence))
        
        chunks = []
        current_chunk = []
        current_size = 0
        
        for sentence in sentences:
            sentence_size = len(sentence) + 1
            if current_size + sentence_size > self.chunk_size and current_chunk:
                chunks.append('\n'.join(current_chunk))
                # オーバーラップの適用
                if self.overlap_size > 0 and len(current_chunk) > 1:
                    overlap_sentences = current_chunk[-self.overlap_size:]
                    current_chunk = overlap_sentences
                    current_size = sum(len(s) + 1 for s in overlap_sentences)
                else:
                    current_chunk = []
                    current_size = 0
            current_chunk.append(sentence)
            current_size += sentence_size
        
        if current_chunk:
            chunks.append('\n'.join(current_chunk))
        
        return chunks

    def _apply_overlap(self, chunks: List[str]) -> List[str]:
        """
        Apply overlap to chunks
        チャンクにオーバーラップを適用
        Args:
            chunks: List of text chunks
            chunks: チャンク化されたテキストのリスト
        Returns:
            List of overlapped chunks
            オーバーラップ適用後のチャンクリスト
        """
        if self.overlap_size <= 0 or len(chunks) <= 1:
            return chunks
        overlapped = []
        for i, chunk in enumerate(chunks):
            if i == 0:
                overlapped.append(chunk)
            else:
                prev = overlapped[-1]
                overlap = prev[-self.overlap_size:] if len(prev) > self.overlap_size else prev
                overlapped.append(overlap + chunk)
        return overlapped 


"""
Evaluator - Metrics Aggregation

テスト結果を集約し、システム性能のメトリクスを計算するDocumentProcessor。
複数の評価指標を統合して包括的な評価レポートを生成します。
"""

from typing import List, Dict, Any, Optional, Union
from pydantic import BaseModel, Field
from dataclasses import field
import json
import statistics
from pathlib import Path
from dataclasses import dataclass

from ..document_processor import DocumentProcessor, DocumentProcessorConfig
from ..models.document import Document


@dataclass
class MetricResult:
    """メトリクス計算結果"""
    name: str
    value: float
    unit: str = ""
    description: str = ""
    threshold: Optional[float] = None
    passed: Optional[bool] = None


class EvaluationMetrics(BaseModel):
    """評価メトリクス定義"""
    
    accuracy: float = 0.0
    precision: float = 0.0
    recall: float = 0.0
    f1_score: float = 0.0
    average_confidence: float = 0.0
    average_response_time: float = 0.0
    source_accuracy: float = 0.0
    coverage: float = 0.0
    consistency: float = 0.0
    user_satisfaction: float = 0.0


class CategoryMetrics(BaseModel):
    """カテゴリ別メトリクス"""
    
    category: str
    total_queries: int
    successful_queries: int
    success_rate: float
    average_confidence: float
    average_response_time: float
    common_failures: List[str] = field(default_factory=list)


class EvaluatorConfig(DocumentProcessorConfig):
    """Evaluator設定"""
    
    include_category_analysis: bool = True
    include_temporal_analysis: bool = False
    include_failure_analysis: bool = True
    confidence_threshold: float = 0.7
    response_time_threshold: float = 2.0
    accuracy_threshold: float = 0.8
    output_format: str = "markdown"  # "markdown", "json", "html"
    
    # メトリクス重み
    metric_weights: Dict[str, float] = field(default_factory=lambda: {
        "accuracy": 0.3,
        "response_time": 0.2,
        "confidence": 0.2,
        "source_accuracy": 0.15,
        "coverage": 0.15
    })


class Evaluator(DocumentProcessor):
    """
    メトリクス集約者
    
    テスト結果ドキュメントを処理し、システム性能の包括的な評価を行います。
    """
    
    @classmethod
    def get_config_class(cls):
        return EvaluatorConfig
    
    def __init__(self, config: EvaluatorConfig):
        super().__init__(config)
        self.evaluation_results: List[Dict[str, Any]] = []
        self.computed_metrics: Optional[EvaluationMetrics] = None
        self.category_metrics: Dict[str, CategoryMetrics] = {}
    
    def process(self, document: Document) -> List[Document]:
        """
        テスト結果ドキュメントを評価メトリクスに変換
        
        Args:
            document: TestSuiteからの結果ドキュメント
            
        Returns:
            List[Document]: 評価メトリクスドキュメント
        """
        # ドキュメントがテスト結果かチェック
        if not self._is_test_result_document(document):
            return [document]  # テスト結果でなければそのまま通す
        
        # テスト結果を解析
        test_results = self._parse_test_results(document)
        self.evaluation_results.extend(test_results)
        
        # メトリクスを計算
        metrics = self._compute_metrics(test_results)
        
        # カテゴリ別分析
        if self.config.include_category_analysis:
            category_analysis = self._analyze_by_category(test_results)
            self.category_metrics.update(category_analysis)
        
        # 失敗分析
        failure_analysis = {}
        if self.config.include_failure_analysis:
            failure_analysis = self._analyze_failures(test_results)
        
        # 評価レポートドキュメントを作成
        report_doc = Document(
            id=f"evaluation_report_{document.id}",
            content=self._format_evaluation_report(metrics, category_analysis, failure_analysis),
            metadata={
                "processing_stage": "evaluation",
                "source_document_id": document.id,
                "metrics_computed": len(metrics.__dict__),
                "categories_analyzed": len(category_analysis) if category_analysis else 0,
                "overall_score": self._compute_overall_score(metrics),
                "recommendations": self._generate_recommendations(metrics)
            }
        )
        
        return [report_doc]
    
    def _is_test_result_document(self, document: Document) -> bool:
        """ドキュメントがテスト結果かチェック"""
        
        processing_stage = document.metadata.get("processing_stage", "")
        return processing_stage in ["test_execution", "test_results"]
    
    def _parse_test_results(self, document: Document) -> List[Dict[str, Any]]:
        """テスト結果ドキュメントから構造化データを抽出"""
        
        results = []
        content = document.content
        
        # コンテンツからテスト結果を解析（簡易実装）
        lines = content.split('\n')
        current_test = {}
        
        for line in lines:
            line = line.strip()
            
            if line.startswith("## ✅ PASS") or line.startswith("## ❌ FAIL"):
                if current_test:
                    results.append(current_test)
                    current_test = {}
                
                current_test["passed"] = "✅ PASS" in line
                current_test["test_id"] = line.split()[-1] if len(line.split()) > 2 else "unknown"
            
            elif line.startswith("**Query**:"):
                current_test["query"] = line.replace("**Query**:", "").strip()
            
            elif line.startswith("**Confidence**:"):
                try:
                    current_test["confidence"] = float(line.replace("**Confidence**:", "").strip())
                except:
                    current_test["confidence"] = 0.0
            
            elif line.startswith("**Processing Time**:"):
                try:
                    time_str = line.replace("**Processing Time**:", "").replace("s", "").strip()
                    current_test["processing_time"] = float(time_str)
                except:
                    current_test["processing_time"] = 0.0
            
            elif line.startswith("**Sources Found**:"):
                try:
                    current_test["sources_found"] = int(line.replace("**Sources Found**:", "").strip())
                except:
                    current_test["sources_found"] = 0
        
        # 最後のテストを追加
        if current_test:
            results.append(current_test)
        
        # メタデータから統計情報も抽出
        doc_metadata = document.metadata
        if "tests_run" in doc_metadata:
            # メタデータベースの統計を結果に統合
            for result in results:
                result["document_id"] = doc_metadata.get("source_document_id", "unknown")
                result["success_rate"] = doc_metadata.get("success_rate", 0.0)
        
        return results
    
    def _compute_metrics(self, test_results: List[Dict[str, Any]]) -> EvaluationMetrics:
        """テスト結果からメトリクスを計算"""
        
        if not test_results:
            return EvaluationMetrics()
        
        # 基本統計
        total_tests = len(test_results)
        passed_tests = sum(1 for r in test_results if r.get("passed", False))
        
        # 各メトリクスを計算
        accuracy = passed_tests / total_tests if total_tests > 0 else 0.0
        
        confidences = [r.get("confidence", 0.0) for r in test_results]
        average_confidence = statistics.mean(confidences) if confidences else 0.0
        
        response_times = [r.get("processing_time", 0.0) for r in test_results]
        average_response_time = statistics.mean(response_times) if response_times else 0.0
        
        # 適合率・再現率の計算（簡易版）
        true_positives = sum(1 for r in test_results 
                           if r.get("passed", False) and r.get("sources_found", 0) > 0)
        false_positives = sum(1 for r in test_results 
                            if not r.get("passed", False) and r.get("sources_found", 0) > 0)
        false_negatives = sum(1 for r in test_results 
                            if r.get("passed", False) and r.get("sources_found", 0) == 0)
        
        precision = true_positives / (true_positives + false_positives) if (true_positives + false_positives) > 0 else 0.0
        recall = true_positives / (true_positives + false_negatives) if (true_positives + false_negatives) > 0 else 0.0
        f1_score = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0.0
        
        # ソース精度
        source_accuracy = sum(1 for r in test_results if r.get("sources_found", 0) > 0) / total_tests
        
        # カバレッジ（テストの多様性）
        unique_queries = len(set(r.get("query", "") for r in test_results))
        coverage = unique_queries / total_tests if total_tests > 0 else 0.0
        
        # 一貫性（同様のクエリでの性能ばらつき）
        confidence_std = statistics.stdev(confidences) if len(confidences) > 1 else 0.0
        consistency = max(0.0, 1.0 - (confidence_std / average_confidence)) if average_confidence > 0 else 0.0
        
        metrics = EvaluationMetrics(
            accuracy=accuracy,
            precision=precision,
            recall=recall,
            f1_score=f1_score,
            average_confidence=average_confidence,
            average_response_time=average_response_time,
            source_accuracy=source_accuracy,
            coverage=coverage,
            consistency=consistency,
            user_satisfaction=self._estimate_user_satisfaction(test_results)
        )
        
        self.computed_metrics = metrics
        return metrics
    
    def _estimate_user_satisfaction(self, test_results: List[Dict[str, Any]]) -> float:
        """ユーザー満足度を推定"""
        
        # 複数要因から満足度を推定
        satisfaction_factors = []
        
        for result in test_results:
            # 基本的な成功
            success_factor = 1.0 if result.get("passed", False) else 0.0
            
            # 信頼度要因
            confidence = result.get("confidence", 0.0)
            confidence_factor = min(confidence / self.config.confidence_threshold, 1.0)
            
            # 応答時間要因
            response_time = result.get("processing_time", 0.0)
            time_factor = max(0.0, 1.0 - (response_time / self.config.response_time_threshold))
            
            # 重み付き平均
            overall_satisfaction = (
                success_factor * 0.5 + 
                confidence_factor * 0.3 + 
                time_factor * 0.2
            )
            
            satisfaction_factors.append(overall_satisfaction)
        
        return statistics.mean(satisfaction_factors) if satisfaction_factors else 0.0
    
    def _analyze_by_category(self, test_results: List[Dict[str, Any]]) -> Dict[str, CategoryMetrics]:
        """カテゴリ別の分析を実行"""
        
        category_data = {}
        
        # カテゴリ別にテスト結果をグループ化
        for result in test_results:
            # カテゴリの推定（クエリ内容から）
            category = self._categorize_result(result)
            
            if category not in category_data:
                category_data[category] = []
            
            category_data[category].append(result)
        
        # 各カテゴリのメトリクスを計算
        category_metrics = {}
        
        for category, results in category_data.items():
            total = len(results)
            successful = sum(1 for r in results if r.get("passed", False))
            
            confidences = [r.get("confidence", 0.0) for r in results]
            times = [r.get("processing_time", 0.0) for r in results]
            
            # 共通的な失敗パターンを特定
            failed_results = [r for r in results if not r.get("passed", False)]
            common_failures = self._identify_common_failures(failed_results)
            
            metrics = CategoryMetrics(
                category=category,
                total_queries=total,
                successful_queries=successful,
                success_rate=successful / total if total > 0 else 0.0,
                average_confidence=statistics.mean(confidences) if confidences else 0.0,
                average_response_time=statistics.mean(times) if times else 0.0,
                common_failures=common_failures
            )
            
            category_metrics[category] = metrics
        
        return category_metrics
    
    def _categorize_result(self, result: Dict[str, Any]) -> str:
        """テスト結果をカテゴリ分類"""
        
        query = result.get("query", "").lower()
        
        if any(word in query for word in ["とは", "何ですか", "どのような"]):
            return "definition"
        elif any(word in query for word in ["方法", "手順", "やり方"]):
            return "how_to"
        elif any(word in query for word in ["なぜ", "理由", "原因"]):
            return "why"
        elif any(word in query for word in ["比較", "違い", "差"]):
            return "comparison"
        elif any(word in query for word in ["天気", "料理", "スポーツ"]):
            return "negative"
        else:
            return "general"
    
    def _identify_common_failures(self, failed_results: List[Dict[str, Any]]) -> List[str]:
        """失敗パターンの共通点を特定"""
        
        failure_patterns = []
        
        if not failed_results:
            return failure_patterns
        
        # 低信頼度
        low_confidence_count = sum(1 for r in failed_results if r.get("confidence", 1.0) < 0.3)
        if low_confidence_count > len(failed_results) * 0.5:
            failure_patterns.append("低信頼度")
        
        # 長い応答時間
        slow_response_count = sum(1 for r in failed_results if r.get("processing_time", 0.0) > 2.0)
        if slow_response_count > len(failed_results) * 0.3:
            failure_patterns.append("応答時間遅延")
        
        # ソース不足
        no_source_count = sum(1 for r in failed_results if r.get("sources_found", 1) == 0)
        if no_source_count > len(failed_results) * 0.4:
            failure_patterns.append("関連ソース不足")
        
        return failure_patterns
    
    def _analyze_failures(self, test_results: List[Dict[str, Any]]) -> Dict[str, Any]:
        """詳細な失敗分析"""
        
        failed_results = [r for r in test_results if not r.get("passed", False)]
        
        if not failed_results:
            return {"total_failures": 0, "failure_rate": 0.0}
        
        failure_analysis = {
            "total_failures": len(failed_results),
            "failure_rate": len(failed_results) / len(test_results),
            "common_patterns": self._identify_common_failures(failed_results),
            "failure_categories": {},
            "improvement_suggestions": []
        }
        
        # カテゴリ別失敗率
        for category in ["definition", "how_to", "why", "comparison", "negative", "general"]:
            category_results = [r for r in test_results if self._categorize_result(r) == category]
            category_failures = [r for r in category_results if not r.get("passed", False)]
            
            if category_results:
                failure_analysis["failure_categories"][category] = {
                    "total": len(category_results),
                    "failures": len(category_failures),
                    "failure_rate": len(category_failures) / len(category_results)
                }
        
        # 改善提案
        if "低信頼度" in failure_analysis["common_patterns"]:
            failure_analysis["improvement_suggestions"].append("モデルの信頼度校正を改善")
        
        if "応答時間遅延" in failure_analysis["common_patterns"]:
            failure_analysis["improvement_suggestions"].append("検索インデックスの最適化")
        
        if "関連ソース不足" in failure_analysis["common_patterns"]:
            failure_analysis["improvement_suggestions"].append("コーパスの拡充または検索戦略の改善")
        
        return failure_analysis
    
    def _compute_overall_score(self, metrics: EvaluationMetrics) -> float:
        """重み付きメトリクスから総合スコアを計算"""
        
        weights = self.config.metric_weights
        
        score = (
            metrics.accuracy * weights.get("accuracy", 0.3) +
            (1.0 - min(metrics.average_response_time / self.config.response_time_threshold, 1.0)) * weights.get("response_time", 0.2) +
            metrics.average_confidence * weights.get("confidence", 0.2) +
            metrics.source_accuracy * weights.get("source_accuracy", 0.15) +
            metrics.coverage * weights.get("coverage", 0.15)
        )
        
        return min(max(score, 0.0), 1.0)  # 0-1の範囲に正規化
    
    def _generate_recommendations(self, metrics: EvaluationMetrics) -> List[str]:
        """メトリクスに基づく改善提案を生成"""
        
        recommendations = []
        
        if metrics.accuracy < self.config.accuracy_threshold:
            recommendations.append("システム精度の改善が必要です")
        
        if metrics.average_response_time > self.config.response_time_threshold:
            recommendations.append("応答時間の最適化を検討してください")
        
        if metrics.average_confidence < self.config.confidence_threshold:
            recommendations.append("モデルの信頼度校正を改善してください")
        
        if metrics.coverage < 0.7:
            recommendations.append("テストケースの多様性を増やしてください")
        
        if metrics.consistency < 0.8:
            recommendations.append("結果の一貫性向上が必要です")
        
        return recommendations
    
    def _format_evaluation_report(
        self, 
        metrics: EvaluationMetrics, 
        category_analysis: Dict[str, CategoryMetrics],
        failure_analysis: Dict[str, Any]
    ) -> str:
        """評価レポートをフォーマット"""
        
        if self.config.output_format == "json":
            return self._format_json_report(metrics, category_analysis, failure_analysis)
        else:
            return self._format_markdown_report(metrics, category_analysis, failure_analysis)
    
    def _format_markdown_report(
        self, 
        metrics: EvaluationMetrics, 
        category_analysis: Dict[str, CategoryMetrics],
        failure_analysis: Dict[str, Any]
    ) -> str:
        """Markdown形式の評価レポート"""
        
        lines = ["# RAGシステム評価レポート\n"]
        
        # 総合スコア
        overall_score = self._compute_overall_score(metrics)
        lines.append(f"## 📊 総合評価: {overall_score:.2f}/1.00")
        
        if overall_score >= 0.8:
            lines.append("🌟 **評価: 優秀** - システムは高品質で安定しています")
        elif overall_score >= 0.6:
            lines.append("👍 **評価: 良好** - 一部改善の余地があります")
        elif overall_score >= 0.4:
            lines.append("📈 **評価: 改善必要** - 重要な問題があります")
        else:
            lines.append("🔧 **評価: 要大幅改善** - システムの根本的な見直しが必要")
        
        lines.append("")
        
        # 主要メトリクス
        lines.append("## 📈 主要メトリクス")
        lines.append(f"- **精度 (Accuracy)**: {metrics.accuracy:.1%}")
        lines.append(f"- **適合率 (Precision)**: {metrics.precision:.1%}")
        lines.append(f"- **再現率 (Recall)**: {metrics.recall:.1%}")
        lines.append(f"- **F1スコア**: {metrics.f1_score:.3f}")
        lines.append(f"- **平均信頼度**: {metrics.average_confidence:.3f}")
        lines.append(f"- **平均応答時間**: {metrics.average_response_time:.3f}秒")
        lines.append(f"- **ソース精度**: {metrics.source_accuracy:.1%}")
        lines.append(f"- **カバレッジ**: {metrics.coverage:.1%}")
        lines.append(f"- **一貫性**: {metrics.consistency:.1%}")
        lines.append(f"- **推定ユーザー満足度**: {metrics.user_satisfaction:.1%}")
        lines.append("")
        
        # カテゴリ別分析
        if category_analysis:
            lines.append("## 📊 カテゴリ別分析")
            for category, cat_metrics in category_analysis.items():
                lines.append(f"### {category.title()}")
                lines.append(f"- 総クエリ数: {cat_metrics.total_queries}")
                lines.append(f"- 成功率: {cat_metrics.success_rate:.1%}")
                lines.append(f"- 平均信頼度: {cat_metrics.average_confidence:.3f}")
                lines.append(f"- 平均応答時間: {cat_metrics.average_response_time:.3f}秒")
                if cat_metrics.common_failures:
                    lines.append(f"- 共通失敗パターン: {', '.join(cat_metrics.common_failures)}")
                lines.append("")
        
        # 失敗分析
        if failure_analysis.get("total_failures", 0) > 0:
            lines.append("## ⚠️ 失敗分析")
            lines.append(f"- 総失敗数: {failure_analysis['total_failures']}")
            lines.append(f"- 失敗率: {failure_analysis['failure_rate']:.1%}")
            
            if failure_analysis.get("common_patterns"):
                lines.append(f"- 共通パターン: {', '.join(failure_analysis['common_patterns'])}")
            
            if failure_analysis.get("improvement_suggestions"):
                lines.append("### 🔧 改善提案")
                for suggestion in failure_analysis["improvement_suggestions"]:
                    lines.append(f"- {suggestion}")
                lines.append("")
        
        # 閾値との比較
        lines.append("## 🎯 閾値との比較")
        accuracy_status = "✅" if metrics.accuracy >= self.config.accuracy_threshold else "❌"
        response_status = "✅" if metrics.average_response_time <= self.config.response_time_threshold else "❌"
        confidence_status = "✅" if metrics.average_confidence >= self.config.confidence_threshold else "❌"
        
        lines.append(f"- 精度閾値 ({self.config.accuracy_threshold:.1%}): {accuracy_status}")
        lines.append(f"- 応答時間閾値 ({self.config.response_time_threshold}秒): {response_status}")
        lines.append(f"- 信頼度閾値 ({self.config.confidence_threshold:.1f}): {confidence_status}")
        
        return "\n".join(lines)
    
    def _format_json_report(
        self, 
        metrics: EvaluationMetrics, 
        category_analysis: Dict[str, CategoryMetrics],
        failure_analysis: Dict[str, Any]
    ) -> str:
        """JSON形式の評価レポート"""
        
        report = {
            "overall_score": self._compute_overall_score(metrics),
            "metrics": metrics.dict(),
            "category_analysis": {k: v.dict() for k, v in category_analysis.items()},
            "failure_analysis": failure_analysis,
            "recommendations": self._generate_recommendations(metrics),
            "thresholds": {
                "accuracy": self.config.accuracy_threshold,
                "response_time": self.config.response_time_threshold,
                "confidence": self.config.confidence_threshold
            }
        }
        
        return json.dumps(report, ensure_ascii=False, indent=2)
    
    def get_summary_metrics(self) -> Dict[str, float]:
        """サマリーメトリクスを取得"""
        
        if not self.computed_metrics:
            return {}
        
        return {
            "overall_score": self._compute_overall_score(self.computed_metrics),
            "accuracy": self.computed_metrics.accuracy,
            "f1_score": self.computed_metrics.f1_score,
            "average_confidence": self.computed_metrics.average_confidence,
            "average_response_time": self.computed_metrics.average_response_time,
            "user_satisfaction": self.computed_metrics.user_satisfaction
        }


import html.parser
import re
from pathlib import Path
from typing import List, Optional, Dict, Any, Iterator
from refinire_rag.models.document import Document
from refinire_rag.loader.loader import Loader

class HTMLParser(html.parser.HTMLParser):
    """
    Custom HTML parser to extract text content.
    テキストコンテンツを抽出するためのカスタムHTMLパーサー。
    """
    def __init__(self):
        """
        Initialize the HTML parser.
        HTMLパーサーを初期化する。
        """
        super().__init__()
        self.text_parts = []
        self.current_tag = None
        self.ignore_tags = {'script', 'style', 'meta', 'link', 'head'}
        self.block_tags = {'p', 'div', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'li', 'td', 'th'}

    def handle_starttag(self, tag: str, attrs: List[tuple[str, Optional[str]]]) -> None:
        """
        Handle the start of an HTML tag.
        HTMLタグの開始を処理する。

        Args:
            tag (str): The tag name.
                      タグ名。
            attrs (List[tuple[str, Optional[str]]]): The tag attributes.
                                                    タグの属性。
        """
        self.current_tag = tag.lower()
        if self.current_tag in self.block_tags:
            self.text_parts.append('\n')

    def handle_endtag(self, tag: str) -> None:
        """
        Handle the end of an HTML tag.
        HTMLタグの終了を処理する。

        Args:
            tag (str): The tag name.
                      タグ名。
        """
        if tag.lower() in self.block_tags:
            self.text_parts.append('\n')
        self.current_tag = None

    def handle_data(self, data: str) -> None:
        """
        Handle the text content of an HTML tag.
        HTMLタグのテキストコンテンツを処理する。

        Args:
            data (str): The text content.
                       テキストコンテンツ。
        """
        if self.current_tag not in self.ignore_tags:
            # Remove extra whitespace
            # 余分な空白を削除
            data = re.sub(r'\s+', ' ', data.strip())
            if data:
                self.text_parts.append(data)

    def get_text(self) -> str:
        """
        Get the extracted text content.
        抽出されたテキストコンテンツを取得する。

        Returns:
            str: The extracted text content.
                 抽出されたテキストコンテンツ。
        """
        # Join text parts and remove extra whitespace
        # テキストパーツを結合し、余分な空白を削除
        text = ' '.join(self.text_parts)
        return re.sub(r'\s+', ' ', text).strip()

class HTMLLoader(Loader):
    """
    Load HTML files and convert their content into Documents.
    HTMLファイルを読み込み、そのコンテンツをDocumentに変換する。
    """
    def __init__(self, encoding: str = 'utf-8'):
        """
        Initialize the HTMLLoader.
        HTMLLoaderを初期化する。

        Args:
            encoding (str): The encoding to use when reading the HTML file.
                           HTMLファイルを読み込む際に使用するエンコーディング。
        """
        super().__init__()
        self.encoding = encoding

    def process(self, documents: List[Document]) -> Iterator[Document]:
        """
        Process the documents by loading HTML files and converting their content into Documents.
        ドキュメントを処理し、HTMLファイルを読み込んでそのコンテンツをDocumentに変換する。

        Args:
            documents (List[Document]): List of documents containing file paths.
                                       ファイルパスを含むドキュメントのリスト。

        Yields:
            Document: A document containing the extracted text content.
                     抽出されたテキストコンテンツを含むドキュメント。
        """
        for doc in documents:
            file_path = doc.metadata.get('file_path')
            if not file_path:
                continue

            try:
                with open(file_path, 'r', encoding=self.encoding) as f:
                    html_content = f.read()

                # Parse HTML content
                # HTMLコンテンツを解析
                parser = HTMLParser()
                parser.feed(html_content)
                text_content = parser.get_text()

                # Create metadata
                # メタデータを作成
                metadata = doc.metadata.copy()
                metadata.update({
                    'content_type': 'html',
                    'file_encoding': self.encoding
                })

                # Create a new document
                # 新しいドキュメントを作成
                yield Document(
                    id=doc.id,
                    content=text_content,
                    metadata=metadata
                )

            except FileNotFoundError:
                raise FileNotFoundError(f"HTML file not found: {file_path}")
            except Exception as e:
                raise Exception(f"Error processing HTML file {file_path}: {str(e)}") 


"""
Evaluation module for RAG system assessment

RAGシステム評価用モジュール

This module provides comprehensive evaluation capabilities including:
- Base evaluator classes for different evaluation types
- Specific metric implementations (QuestEval, BLEU, ROUGE, LLM Judge)
- Composite evaluation for combining multiple metrics
"""

from .base_evaluator import (
    BaseEvaluator,
    ReferenceBasedEvaluator,
    ReferenceFreeEvaluator,
    CompositeEvaluator,
    BaseEvaluatorConfig,
    EvaluationScore
)

from .questeval_evaluator import QuestEvalEvaluator, QuestEvalConfig
from .bleu_evaluator import BLEUEvaluator, BLEUConfig
from .rouge_evaluator import ROUGEEvaluator, ROUGEConfig
from .llm_judge_evaluator import LLMJudgeEvaluator, LLMJudgeConfig

__all__ = [
    # Base classes
    "BaseEvaluator",
    "ReferenceBasedEvaluator", 
    "ReferenceFreeEvaluator",
    "CompositeEvaluator",
    "BaseEvaluatorConfig",
    "EvaluationScore",
    
    # Specific evaluators
    "QuestEvalEvaluator",
    "QuestEvalConfig",
    "BLEUEvaluator",
    "BLEUConfig", 
    "ROUGEEvaluator",
    "ROUGEConfig",
    "LLMJudgeEvaluator",
    "LLMJudgeConfig"
]


def create_comprehensive_evaluator(
    enable_questeval: bool = True,
    enable_bleu: bool = True,
    enable_rouge: bool = True,
    enable_llm_judge: bool = True,
    questeval_config: QuestEvalConfig = None,
    bleu_config: BLEUConfig = None,
    rouge_config: ROUGEConfig = None,
    llm_judge_config: LLMJudgeConfig = None
) -> CompositeEvaluator:
    """
    Create a comprehensive evaluator combining multiple metrics
    
    複数のメトリクスを組み合わせた包括的評価器を作成
    
    Args:
        enable_questeval: Enable QuestEval evaluator
                         QuestEval評価器を有効化
        enable_bleu: Enable BLEU evaluator
                    BLEU評価器を有効化
        enable_rouge: Enable ROUGE evaluator
                     ROUGE評価器を有効化
        enable_llm_judge: Enable LLM Judge evaluator
                         LLM Judge評価器を有効化
        questeval_config: Custom QuestEval configuration
                         カスタムQuestEval設定
        bleu_config: Custom BLEU configuration
                    カスタムBLEU設定
        rouge_config: Custom ROUGE configuration
                     カスタムROUGE設定
        llm_judge_config: Custom LLM Judge configuration
                         カスタムLLM Judge設定
    
    Returns:
        CompositeEvaluator: Combined evaluator with all enabled metrics
                           有効化されたすべてのメトリクスを含む複合評価器
    """
    evaluators = []
    
    if enable_questeval:
        config = questeval_config or QuestEvalConfig()
        evaluators.append(QuestEvalEvaluator(config))
    
    if enable_bleu:
        config = bleu_config or BLEUConfig()
        evaluators.append(BLEUEvaluator(config))
    
    if enable_rouge:
        config = rouge_config or ROUGEConfig()
        evaluators.append(ROUGEEvaluator(config))
    
    if enable_llm_judge:
        config = llm_judge_config or LLMJudgeConfig()
        evaluators.append(LLMJudgeEvaluator(config))
    
    return CompositeEvaluator(evaluators)


def create_quick_evaluator(evaluation_type: str = "comprehensive") -> BaseEvaluator:
    """
    Create a quick evaluator for common use cases
    
    一般的な用途向けのクイック評価器を作成
    
    Args:
        evaluation_type: Type of evaluation to perform
                        実行する評価の種類
                        Options: "comprehensive", "lexical", "semantic", "llm_only"
    
    Returns:
        BaseEvaluator: Configured evaluator
                      設定済み評価器
    """
    if evaluation_type == "comprehensive":
        return create_comprehensive_evaluator()
    
    elif evaluation_type == "lexical":
        # Focus on lexical overlap metrics
        return create_comprehensive_evaluator(
            enable_questeval=False,
            enable_bleu=True,
            enable_rouge=True,
            enable_llm_judge=False
        )
    
    elif evaluation_type == "semantic":
        # Focus on semantic evaluation
        return create_comprehensive_evaluator(
            enable_questeval=True,
            enable_bleu=False,
            enable_rouge=False,
            enable_llm_judge=True
        )
    
    elif evaluation_type == "llm_only":
        # Use only LLM-based evaluation
        return LLMJudgeEvaluator(LLMJudgeConfig())
    
    else:
        raise ValueError(f"Unknown evaluation type: {evaluation_type}")


"""
CorpusManager - Document corpus construction and management

Simplified CorpusManager with core functionality for document import and rebuild.
"""

import logging
import time
from typing import List, Optional, Dict, Any
from dataclasses import dataclass
from pathlib import Path

from ..processing.normalizer import NormalizerConfig
from ..processing.chunker import ChunkingConfig
from ..loader.document_store_loader import DocumentStoreLoader, DocumentLoadConfig, LoadStrategy
from ..models.document import Document

logger = logging.getLogger(__name__)


@dataclass
class CorpusStats:
    """Statistics for corpus building operations"""
    total_files_processed: int = 0
    total_documents_created: int = 0
    total_chunks_created: int = 0
    total_processing_time: float = 0.0
    pipeline_stages_executed: int = 0
    documents_by_stage: Dict[str, int] = None
    errors_encountered: int = 0
    
    def __post_init__(self):
        if self.documents_by_stage is None:
            self.documents_by_stage = {}


class CorpusManager:
    """Document corpus construction and management system
    
    Simplified corpus manager that provides:
    - Document import from folders with incremental loading
    - Corpus rebuild from original documents using existing knowledge artifacts
    - Corpus clearing functionality
    
    Environment Variables:
    - REFINIRE_DIR: Base directory for Refinire files (default: './refinire')
    - REFINIRE_RAG_CORPUS_STORE: Corpus store type (default: 'sqlite')
    
    File Naming Convention:
    - Tracking file: {corpus_name}_track.json
    - Dictionary file: {corpus_name}_dictionary.md
    - Knowledge graph file: {corpus_name}_knowledge_graph.md
    """
    
    def __init__(self, document_store, vector_store, config: Optional[Dict[str, Any]] = None):
        """Initialize CorpusManager
        
        Args:
            document_store: DocumentStore for document persistence
            vector_store: VectorStore for vector persistence
            config: Optional global configuration
        """
        self.document_store = document_store
        self.vector_store = vector_store
        self.config = config or {}
        self.stats = CorpusStats()
        
        logger.info(f"Initialized CorpusManager with DocumentStore: {type(document_store).__name__}, "
                   f"VectorStore: {type(vector_store).__name__}")
    
    @staticmethod
    def _get_refinire_rag_dir() -> Path:
        """Get REFINIRE_DIR/rag directory from environment variable or default
        
        環境変数REFINIRE_DIRまたはデフォルトディレクトリを取得し、/ragサブディレクトリを使用
        
        Returns:
            Path to the REFINIRE_DIR/rag directory
        """
        import os
        from pathlib import Path
        
        # Check REFINIRE_DIR environment variable first
        base_dir = os.getenv("REFINIRE_DIR", "./refinire")
        rag_path = Path(base_dir) / "rag"
        rag_path.mkdir(parents=True, exist_ok=True)
        
        return rag_path
    
    @staticmethod
    def _get_corpus_file_path(corpus_name: str, file_type: str, custom_dir: Optional[str] = None) -> Path:
        """Get corpus-specific file path
        
        コーパス固有のファイルパスを取得
        
        Args:
            corpus_name: Name of the corpus
            file_type: Type of file ('track', 'dictionary', 'knowledge_graph')
            custom_dir: Custom directory path (overrides default)
            
        Returns:
            Path to the corpus file
        """
        if custom_dir:
            base_dir = Path(custom_dir)
        else:
            base_dir = CorpusManager._get_refinire_rag_dir()
        
        base_dir.mkdir(parents=True, exist_ok=True)
        
        # File naming convention: [corpus_name]_[file_type].[ext]
        if file_type == "track":
            filename = f"{corpus_name}_track.json"
        elif file_type == "dictionary":
            filename = f"{corpus_name}_dictionary.md"
        elif file_type == "knowledge_graph":
            filename = f"{corpus_name}_knowledge_graph.md"
        else:
            raise ValueError(f"Unknown file type: {file_type}")
        
        return base_dir / filename
    
    @staticmethod
    def _get_default_output_directory(env_var_name: str, subdir: str) -> Path:
        """Get default output directory from environment variable or .refinire/
        
        環境変数またはデフォルト.refinire/ディレクトリから出力ディレクトリを取得
        
        Args:
            env_var_name: Environment variable name to check
            subdir: Subdirectory name under .refinire/
            
        Returns:
            Path to the output directory
        """
        import os
        from pathlib import Path
        
        # Check environment variable first
        env_path = os.getenv(env_var_name)
        if env_path:
            return Path(env_path)
        
        # Fall back to .refinire/subdir in user's home directory
        home_dir = Path.home()
        default_dir = home_dir / ".refinire" / subdir
        default_dir.mkdir(parents=True, exist_ok=True)
        
        return default_dir
    
    def _create_filter_config_from_glob(self, glob_pattern: str) -> Optional['FilterConfig']:
        """Create FilterConfig from glob pattern
        
        globパターンからFilterConfigを作成
        
        Args:
            glob_pattern: Glob pattern (e.g., "**/*.md", "*.{txt,py}")
            
        Returns:
            FilterConfig object or None if no filtering needed
        """
        from ..loader.models.filter_config import FilterConfig
        import re
        
        # Extract file extensions from glob pattern
        extensions = []
        
        # Handle patterns like "**/*.md", "*.txt"
        if '*.' in glob_pattern:
            # Simple extension pattern
            ext_match = re.search(r'\*\.([a-zA-Z0-9]+)', glob_pattern)
            if ext_match:
                extensions.append('.' + ext_match.group(1))
        
        # Handle patterns like "*.{txt,md,py}"
        brace_match = re.search(r'\*\.{([^}]+)}', glob_pattern)
        if brace_match:
            ext_list = brace_match.group(1).split(',')
            extensions.extend(['.' + ext.strip() for ext in ext_list])
        
        # Create filter config if extensions found
        if extensions:
            from ..loader.filters.extension_filter import ExtensionFilter
            extension_filter = ExtensionFilter(include_extensions=extensions)
            return FilterConfig(extension_filter=extension_filter)
        
        # For complex patterns, return None and let glob be handled by the loader
        return None
    
    def import_original_documents(self, 
                                corpus_name: str,
                                directory: str,
                                glob: str = "**/*",
                                use_multithreading: bool = True,
                                force_reload: bool = False,
                                additional_metadata: Optional[Dict[str, Any]] = None,
                                tracking_file_path: Optional[str] = None,
                                create_dictionary: bool = False,
                                create_knowledge_graph: bool = False,
                                dictionary_output_dir: Optional[str] = None,
                                graph_output_dir: Optional[str] = None) -> CorpusStats:
        """Import original documents from specified directory with incremental loading
        
        指定ディレクトリからIncrementalLoaderを使って元文書を取り込み、
        processing_stage: "original"メタデータを自動設定し、オプションで辞書・グラフを作成
        
        Args:
            corpus_name: Name of the corpus (used in metadata and output filenames)
                       コーパス名（メタデータと出力ファイル名に使用）
            directory: Directory path to import from (similar to LangChain DirectoryLoader)
                     取り込み対象ディレクトリパス（LangChain DirectoryLoaderと同様）
            glob: Glob pattern to match files (default: "**/*" for all files recursively)
                ファイルマッチング用のglobパターン（デフォルト: "**/*" 全ファイル再帰的）
            use_multithreading: Whether to use multithreading for file processing
                              ファイル処理にマルチスレッドを使用するか
            force_reload: Force reload all files ignoring incremental cache
                        増分キャッシュを無視してすべてのファイルを強制再読み込み
            additional_metadata: Additional metadata to add to all imported documents
                               すべての取り込み文書に追加する追加メタデータ
            tracking_file_path: Path to store file tracking data for incremental loading
                              増分ローディング用ファイル追跡データの保存パス
            create_dictionary: Whether to create domain dictionary after import
                             取り込み後にドメイン辞書を作成するか
            create_knowledge_graph: Whether to create knowledge graph after import
                                  取り込み後にナレッジグラフを作成するか
            dictionary_output_dir: Directory to save dictionary file (default: env REFINIRE_DICTIONARY_DIR or ~/.refinire/dictionaries)
                                 辞書ファイルの保存ディレクトリ（デフォルト: 環境変数REFINIRE_DICTIONARY_DIRまたは~/.refinire/dictionaries）
            graph_output_dir: Directory to save graph file (default: env REFINIRE_GRAPH_DIR or ~/.refinire/graphs)
                            グラフファイルの保存ディレクトリ（デフォルト: 環境変数REFINIRE_GRAPH_DIRまたは~/.refinire/graphs）
        
        Returns:
            CorpusStats: Import statistics including files processed and documents created
                        処理ファイル数と作成文書数を含む取り込み統計
        
        Example:
            # 基本的な取り込み（全ファイル）
            stats = corpus_manager.import_original_documents(
                corpus_name="product_docs",
                directory="/path/to/docs"
            )
            
            # Markdownファイルのみ取り込み（LangChain風）
            stats = corpus_manager.import_original_documents(
                corpus_name="markdown_docs",
                directory="/path/to/docs",
                glob="**/*.md",
                use_multithreading=True
            )
            
            # 詳細設定での取り込み
            stats = corpus_manager.import_original_documents(
                corpus_name="engineering_docs",
                directory="/path/to/docs",
                glob="**/*.{txt,md,py}",  # 複数拡張子
                use_multithreading=False,
                additional_metadata={"department": "engineering", "project": "rag"},
                tracking_file_path="./import_tracking.json",
                create_dictionary=True,
                create_knowledge_graph=True,
                dictionary_output_dir="./knowledge",
                graph_output_dir="./knowledge"
            )
            # → ./knowledge/engineering_docs_dictionary.md
            # → ./knowledge/engineering_docs_graph.md が作成される
        """
        from ..loader.incremental_directory_loader import IncrementalDirectoryLoader
        from ..metadata.constant_metadata import ConstantMetadata
        from ..loader.models.filter_config import FilterConfig
        from datetime import datetime
        
        stats = CorpusStats()
        
        # Create ConstantMetadata to automatically add processing_stage: "original"
        base_metadata = {
            "processing_stage": "original",
            "import_timestamp": datetime.now().isoformat(),
            "imported_by": "import_original_documents",
            "corpus_name": corpus_name
        }
        
        # Add additional metadata if provided
        if additional_metadata:
            base_metadata.update(additional_metadata)
        
        constant_metadata_processor = ConstantMetadata(base_metadata)
        
        # Set default tracking file path if not provided
        if tracking_file_path is None:
            tracking_file_path = self._get_corpus_file_path(corpus_name, "track")
            logger.info(f"Using default tracking file: {tracking_file_path}")
        else:
            # Convert string to Path if provided as string
            tracking_file_path = Path(tracking_file_path)
        
        # Create filter configuration from glob pattern
        filter_config = None
        if glob != "**/*":
            # Convert glob pattern to filter configuration
            filter_config = self._create_filter_config_from_glob(glob)
        
        try:
            logger.info(f"Importing documents from: {directory}")
            logger.info(f"Using glob pattern: {glob}")
            if use_multithreading:
                logger.warning("Multithreading not yet supported by IncrementalDirectoryLoader, processing sequentially")
            
            # Create incremental loader for the directory
            incremental_loader = IncrementalDirectoryLoader(
                directory_path=directory,
                document_store=self.document_store,
                filter_config=filter_config,
                tracking_file_path=tracking_file_path,
                recursive=True,  # Always recursive since glob pattern can specify depth
                metadata_processors=[constant_metadata_processor]
                # Note: use_multithreading not yet supported by IncrementalDirectoryLoader
            )
            
            # Handle force reload
            if force_reload:
                incremental_loader.file_tracker.clear_tracking_data()
            
            sync_result = incremental_loader.sync_with_store()
            
            # Update statistics
            documents_processed = len(sync_result.added_documents) + len(sync_result.updated_documents)
            stats.total_files_processed += documents_processed
            stats.total_documents_created += documents_processed
            stats.pipeline_stages_executed += 1
            
            # Track by stage
            stage_key = "original"
            if stage_key not in stats.documents_by_stage:
                stats.documents_by_stage[stage_key] = 0
            stats.documents_by_stage[stage_key] += documents_processed
            
            # Track errors
            if sync_result.has_errors:
                stats.errors_encountered += len(sync_result.errors)
            
            logger.info(f"Imported {documents_processed} documents from {directory}")
            
        except Exception as e:
            logger.error(f"Error importing from directory {directory}: {e}")
            stats.errors_encountered += 1
            raise
        
        logger.info(f"Import completed: {stats.total_documents_created} documents from {directory}")
        
        # Create dictionary and/or knowledge graph if requested
        # 辞書・ナレッジグラフ作成（要求された場合）
        if create_dictionary or create_knowledge_graph:
            self._create_knowledge_artifacts(
                corpus_name=corpus_name,
                create_dictionary=create_dictionary,
                create_knowledge_graph=create_knowledge_graph,
                dictionary_output_dir=dictionary_output_dir,
                graph_output_dir=graph_output_dir,
                stats=stats
            )
        
        return stats
    
    def _create_knowledge_artifacts(self,
                                  corpus_name: str,
                                  create_dictionary: bool,
                                  create_knowledge_graph: bool,
                                  dictionary_output_dir: Optional[str],
                                  graph_output_dir: Optional[str],
                                  stats: CorpusStats):
        """Create dictionary and/or knowledge graph from imported documents
        
        取り込み済み文書から辞書・ナレッジグラフを作成
        """
        from ..processing.dictionary_maker import DictionaryMaker, DictionaryMakerConfig
        from ..processing.graph_builder import GraphBuilder, GraphBuilderConfig
        from ..processing.document_pipeline import DocumentPipeline
        from pathlib import Path
        import os
        
        knowledge_stages = []
        
        # Prepare dictionary creation
        if create_dictionary:
            # Use corpus-specific file path with environment variable support
            dict_file_path = self._get_corpus_file_path(corpus_name, "dictionary", dictionary_output_dir)
            
            dict_config = DictionaryMakerConfig(
                dictionary_file_path=str(dict_file_path),
                focus_on_technical_terms=True,
                extract_abbreviations=True,
                detect_expression_variations=True
            )
            knowledge_stages.append(("dictionary", dict_config))
            logger.info(f"Will create dictionary: {dict_file_path}")
        
        # Prepare knowledge graph creation  
        if create_knowledge_graph:
            # Use corpus-specific file path with environment variable support
            graph_file_path = self._get_corpus_file_path(corpus_name, "knowledge_graph", graph_output_dir)
            
            graph_config = GraphBuilderConfig(
                graph_file_path=str(graph_file_path),
                focus_on_important_relationships=True,
                extract_hierarchical_relationships=True,
                extract_causal_relationships=True
            )
            knowledge_stages.append(("graph", graph_config))
            logger.info(f"Will create knowledge graph: {graph_file_path}")
        
        # Execute knowledge extraction stages
        if knowledge_stages:
            try:
                logger.info(f"Creating knowledge artifacts for corpus '{corpus_name}'...")
                
                # Load original documents
                loader_config = DocumentLoadConfig(strategy=LoadStrategy.FILTERED, metadata_filters={"processing_stage": "original"})
                loader = DocumentStoreLoader(self.document_store, load_config=loader_config)
                
                # Create trigger document
                trigger_doc = Document(
                    id="knowledge_creation_trigger",
                    content="",
                    metadata={
                        "trigger_type": "knowledge_creation",
                        "corpus_name": corpus_name
                    }
                )
                
                # Execute each knowledge stage
                for stage_name, stage_config in knowledge_stages:
                    if stage_name == "dictionary":
                        processors = [loader, DictionaryMaker(stage_config)]
                    elif stage_name == "graph":
                        processors = [loader, GraphBuilder(stage_config)]
                    else:
                        continue
                    
                    pipeline = DocumentPipeline(processors)
                    results = pipeline.process_document(trigger_doc)
                    
                    stats.pipeline_stages_executed += 1
                    logger.info(f"Knowledge stage '{stage_name}' completed")
                
                logger.info(f"Knowledge artifacts created successfully for '{corpus_name}'")
                
            except Exception as e:
                logger.error(f"Error creating knowledge artifacts for '{corpus_name}': {e}")
                stats.errors_encountered += 1
    
    def rebuild_corpus_from_original(self,
                                   corpus_name: str,
                                   use_dictionary: bool = True,
                                   use_knowledge_graph: bool = False,
                                   dictionary_file_path: Optional[str] = None,
                                   graph_file_path: Optional[str] = None,
                                   additional_metadata: Optional[Dict[str, Any]] = None,
                                   stage_configs: Optional[Dict[str, Any]] = None) -> CorpusStats:
        """Rebuild corpus from existing original documents using existing knowledge artifacts
        
        既存のoriginalステージ文書から、既存の辞書・ナレッジグラフを利用してコーパスを再構築
        
        Args:
            corpus_name: Name of the corpus for metadata
                       メタデータ用のコーパス名
            use_dictionary: Whether to use existing dictionary for normalization
                          既存辞書を正規化に使用するか
            use_knowledge_graph: Whether to use existing knowledge graph for normalization
                               既存ナレッジグラフを正規化に使用するか
            dictionary_file_path: Path to existing dictionary file to use
                                既存の辞書ファイルパス
            graph_file_path: Path to existing knowledge graph file to use
                           既存のナレッジグラフファイルパス
            additional_metadata: Additional metadata to add during rebuild
                               再構築時に追加するメタデータ
            stage_configs: Configuration for each processing stage
                         各処理ステージの設定
            
        Returns:
            CorpusStats: Rebuild statistics
                        再構築統計
            
        Note:
            This method does NOT create new dictionary or knowledge graph files.
            It uses existing files for normalization if specified.
            このメソッドは新しい辞書やナレッジグラフファイルを作成しません。
            指定された既存ファイルを正規化に使用します。
            
        Example:
            # 基本的な再構築（既存辞書使用）
            stats = corpus_manager.rebuild_corpus_from_original(
                corpus_name="product_docs",
                use_dictionary=True,
                dictionary_file_path="./knowledge/product_docs_dictionary.md",
                additional_metadata={"rebuild_version": "2.0"}
            )
            
            # 辞書+ナレッジグラフ使用での再構築
            stats = corpus_manager.rebuild_corpus_from_original(
                corpus_name="engineering_docs", 
                use_dictionary=True,
                use_knowledge_graph=True,
                dictionary_file_path="./knowledge/engineering_docs_dictionary.md",
                graph_file_path="./knowledge/engineering_docs_graph.md",
                additional_metadata={
                    "rebuild_timestamp": datetime.now().isoformat(),
                    "rebuild_reason": "parameter_tuning"
                },
                stage_configs={
                    "normalizer_config": NormalizerConfig(
                        case_sensitive_replacement=True,
                        whole_word_only=False
                    ),
                    "chunker_config": ChunkingConfig(
                        chunk_size=1024,
                        overlap=100
                    )
                }
            )
        """
        from ..processing.document_pipeline import DocumentPipeline
        from ..processing.normalizer import Normalizer
        from ..processing.chunker import Chunker
        from datetime import datetime
        
        start_time = time.time()
        logger.info(f"Starting corpus rebuild for '{corpus_name}' from original documents")
        
        # Check if original documents exist
        original_docs = list(self._get_documents_by_stage("original"))
        if not original_docs:
            raise ValueError("No original documents found. Please import documents first using import_original_documents()")
        
        logger.info(f"Found {len(original_docs)} original documents to rebuild from")
        
        # Prepare metadata for rebuilt documents
        rebuild_metadata = {
            "rebuild_timestamp": datetime.now().isoformat(),
            "rebuild_corpus_name": corpus_name,
            "rebuilt_from": "original"
        }
        if additional_metadata:
            rebuild_metadata.update(additional_metadata)
        
        # Validate that knowledge files exist if specified
        if use_dictionary:
            if dictionary_file_path:
                dict_path = Path(dictionary_file_path)
                if not dict_path.exists():
                    raise FileNotFoundError(f"Dictionary file not found: {dictionary_file_path}")
                logger.info(f"Using existing dictionary: {dictionary_file_path}")
            else:
                # Try to find corpus-specific dictionary file using environment variables
                default_dict_path = self._get_corpus_file_path(corpus_name, "dictionary")
                if default_dict_path.exists():
                    dictionary_file_path = str(default_dict_path)
                    logger.info(f"Found corpus dictionary: {dictionary_file_path}")
                else:
                    logger.warning(f"No dictionary file specified and corpus dictionary not found: {default_dict_path}")
                    use_dictionary = False
        
        if use_knowledge_graph:
            if graph_file_path:
                graph_path = Path(graph_file_path)
                if not graph_path.exists():
                    raise FileNotFoundError(f"Knowledge graph file not found: {graph_file_path}")
                logger.info(f"Using existing knowledge graph: {graph_file_path}")
            else:
                # Try to find corpus-specific knowledge graph file using environment variables
                default_graph_path = self._get_corpus_file_path(corpus_name, "knowledge_graph")
                if default_graph_path.exists():
                    graph_file_path = str(default_graph_path)
                    logger.info(f"Found corpus knowledge graph: {graph_file_path}")
                else:
                    logger.warning(f"No graph file specified and corpus knowledge graph not found: {default_graph_path}")
                    use_knowledge_graph = False
        
        # Determine stages to execute based on options
        processors = []
        stage_configs = stage_configs or {}
        
        # Load original documents
        loader_config = DocumentLoadConfig(strategy=LoadStrategy.FILTERED, metadata_filters={"processing_stage": "original"})
        loader = DocumentStoreLoader(self.document_store, load_config=loader_config)
        processors.append(loader)
        
        # Add normalization stage if using dictionary or knowledge graph
        if use_dictionary or use_knowledge_graph:
            # Configure normalizer to use existing files
            if not stage_configs.get("normalizer_config"):
                normalizer_config = NormalizerConfig()
                
                # Set dictionary file path if using dictionary
                if use_dictionary and dictionary_file_path:
                    normalizer_config.dictionary_file_path = dictionary_file_path
                    normalizer_config.auto_detect_dictionary_path = False  # Use specified path
                    normalizer_config.skip_if_no_dictionary = False
                
                # Note: Normalizer currently only supports dictionary normalization
                # Knowledge graph integration would need to be implemented separately
                if use_knowledge_graph and graph_file_path:
                    logger.warning("Knowledge graph normalization not yet supported by Normalizer")
                
                stage_configs["normalizer_config"] = normalizer_config
            
            processors.append(Normalizer(stage_configs["normalizer_config"]))
        
        # Add chunking stage
        if not stage_configs.get("chunker_config"):
            stage_configs["chunker_config"] = ChunkingConfig()
        processors.append(Chunker(stage_configs["chunker_config"]))
        
        # Add vectorization stage
        processors.append(self.vector_store)
        
        try:
            # Execute rebuild pipeline
            pipeline = DocumentPipeline(processors)
            
            # Create trigger document
            trigger_doc = Document(
                id="rebuild_trigger",
                content="",
                metadata={
                    "trigger_type": "corpus_rebuild",
                    "corpus_name": corpus_name,
                    **rebuild_metadata
                }
            )
            
            logger.info(f"Executing rebuild pipeline with {len(processors)} processors")
            results = pipeline.process_document(trigger_doc)
            
            # Calculate statistics
            stats = CorpusStats()
            stats.total_documents_created = len(results)
            stats.pipeline_stages_executed = len(processors)
            
            # Count chunks (documents with chunk metadata)
            chunks = [doc for doc in results if doc.metadata.get("processing_stage") == "chunked"]
            stats.total_chunks_created = len(chunks)
            
            # Update timing
            total_time = time.time() - start_time
            stats.total_processing_time = total_time
            
            logger.info(f"Corpus rebuild completed in {total_time:.3f}s for '{corpus_name}': "
                       f"{stats.total_documents_created} documents processed, "
                       f"{stats.total_chunks_created} chunks created")
            
            # Log knowledge artifacts used
            if use_dictionary and dictionary_file_path:
                logger.info(f"Used dictionary: {dictionary_file_path}")
            
            if use_knowledge_graph and graph_file_path:
                logger.info(f"Used knowledge graph: {graph_file_path}")
            
            return stats
            
        except Exception as e:
            logger.error(f"Corpus rebuild failed for '{corpus_name}': {e}")
            raise
    
    def _get_documents_by_stage(self, processing_stage: str) -> List[Document]:
        """Get documents by processing stage
        
        Args:
            processing_stage: Stage to filter by
            
        Returns:
            List of documents in the specified stage
        """
        loader = DocumentStoreLoader(self.document_store, 
                                   load_config=DocumentLoadConfig(strategy=LoadStrategy.FILTERED, 
                                                                 metadata_filters={"processing_stage": processing_stage}))
        
        # Create trigger document
        trigger = Document(id="stage_query", content="", metadata={})
        return loader.process(trigger)
    
    def clear_corpus(self):
        """Clear all documents from the corpus
        
        コーパスからすべての文書を削除
        
        Note:
            This method will remove all documents from both DocumentStore and VectorStore.
            このメソッドはDocumentStoreとVectorStoreの両方からすべての文書を削除します。
        """
        try:
            logger.info("Starting corpus clearing...")
            
            # Clear document store
            if hasattr(self.document_store, 'clear_all_documents'):
                self.document_store.clear_all_documents()
                logger.info("Cleared all documents from DocumentStore")
            else:
                logger.warning("DocumentStore does not support clear_all_documents method")
            
            # Clear vector store
            if hasattr(self.vector_store, 'clear_all_vectors'):
                self.vector_store.clear_all_vectors()
                logger.info("Cleared all vectors from VectorStore")
            else:
                logger.warning("VectorStore does not support clear_all_vectors method")
            
            # Reset stats
            self.stats = CorpusStats()
            
            logger.info("Corpus clearing completed successfully")
            
        except Exception as e:
            logger.error(f"Error clearing corpus: {e}")
            raise


"""
Refinire RAG package.
/ Refinire RAGパッケージ
"""

import importlib
import logging

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

# Get the version from pyproject.toml
# pyproject.tomlからバージョンを取得
try:
    __version__ = version("refinire-rag")
except PackageNotFoundError:
    __version__ = "unknown"

# Import core models
# コアモデルをインポート
from .models import Document, QAPair, EvaluationResult

# Import core interfaces
# コアインターフェースをインポート
from .corpusstore import CorpusStore

# Import implementations
# 実装をインポート
from .corpus_store.sqlite_corpus_store import SQLiteCorpusStore

__all__ = [
    "Document",
    "QAPair",
    "EvaluationResult",
    "CorpusStore",
    "SQLiteCorpusStore",
]


"""
InsightReporter - Threshold-based Interpretation and Reporting

閾値ベースの解釈とレポート生成を行うDocumentProcessor。
評価結果を分析し、実用的なインサイトと推奨事項を生成します。
"""

from typing import List, Dict, Any, Optional, Tuple
from pydantic import BaseModel, Field
from dataclasses import field
from enum import Enum
import statistics
import json
from pathlib import Path
from datetime import datetime

from ..document_processor import DocumentProcessor, DocumentProcessorConfig
from ..models.document import Document


class InsightType(str, Enum):
    """インサイトタイプ"""
    PERFORMANCE = "performance"        # 性能に関するインサイト
    QUALITY = "quality"               # 品質に関するインサイト
    CONSISTENCY = "consistency"       # 一貫性に関するインサイト
    EFFICIENCY = "efficiency"         # 効率性に関するインサイト
    RELIABILITY = "reliability"       # 信頼性に関するインサイト
    SCALABILITY = "scalability"       # スケーラビリティに関するインサイト


class Insight(BaseModel):
    """インサイトモデル"""
    
    id: str
    insight_type: InsightType
    title: str
    description: str
    severity: str  # "critical", "high", "medium", "low", "info"
    confidence: float
    affected_metrics: List[str]
    recommendations: List[str]
    supporting_data: Dict[str, Any] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)


class Threshold(BaseModel):
    """閾値設定"""
    
    metric_name: str
    critical_threshold: Optional[float] = None
    warning_threshold: Optional[float] = None
    target_threshold: Optional[float] = None
    comparison_operator: str = "greater_than"  # "greater_than", "less_than", "equals"


class InsightReporterConfig(DocumentProcessorConfig):
    """InsightReporter設定"""
    
    # インサイト生成設定
    enable_trend_analysis: bool = True
    enable_comparative_analysis: bool = True
    enable_root_cause_analysis: bool = True
    min_confidence_for_insight: float = 0.6
    
    # レポート設定
    include_executive_summary: bool = True
    include_detailed_analysis: bool = True
    include_action_items: bool = True
    report_format: str = "markdown"  # "markdown", "json", "html"


class InsightReporter(DocumentProcessor):
    """
    インサイトレポーター
    
    評価結果を分析し、閾値ベースの解釈とインサイトを生成します。
    """
    
    @classmethod
    def get_config_class(cls):
        return InsightReporterConfig
    
    def __init__(self, config: InsightReporterConfig):
        super().__init__(config)
        self.generated_insights: List[Insight] = []
        self.all_thresholds = self._initialize_thresholds()
        self.historical_data: List[Dict[str, Any]] = []
    
    def process(self, document: Document) -> List[Document]:
        """
        評価結果ドキュメントからインサイトを生成
        
        Args:
            document: 評価結果ドキュメント
            
        Returns:
            List[Document]: インサイトレポートドキュメント
        """
        # ドキュメントが評価結果かチェック
        if not self._is_evaluation_document(document):
            return [document]  # 評価結果でなければそのまま通す
        
        # メトリクスを抽出
        metrics = self._extract_metrics(document)
        
        # 履歴データに追加
        self.historical_data.append({
            "timestamp": datetime.now().isoformat(),
            "document_id": document.id,
            "metrics": metrics
        })
        
        # インサイトを生成
        insights = self._generate_insights(metrics, document)
        self.generated_insights.extend(insights)
        
        # インサイトレポートを作成
        report_doc = Document(
            id=f"insight_report_{document.id}",
            content=self._format_insight_report(insights, metrics),
            metadata={
                "processing_stage": "insight_reporting",
                "source_document_id": document.id,
                "insights_generated": len(insights),
                "critical_insights": len([i for i in insights if i.severity == "critical"]),
                "overall_health_score": self._compute_health_score(metrics),
                "recommendation_count": sum(len(i.recommendations) for i in insights)
            }
        )
        
        return [report_doc]
    
    def _initialize_thresholds(self) -> List[Threshold]:
        """閾値設定を初期化"""
        
        # デフォルトの閾値を作成
        thresholds = [
            Threshold(
                metric_name="accuracy",
                critical_threshold=0.5,
                warning_threshold=0.7,
                target_threshold=0.9,
                comparison_operator="greater_than"
            ),
            Threshold(
                metric_name="response_time",
                critical_threshold=5.0,
                warning_threshold=2.0,
                target_threshold=1.0,
                comparison_operator="less_than"
            ),
            Threshold(
                metric_name="confidence",
                critical_threshold=0.3,
                warning_threshold=0.5,
                target_threshold=0.8,
                comparison_operator="greater_than"
            )
        ]
        
        return thresholds
    
    def _is_evaluation_document(self, document: Document) -> bool:
        """ドキュメントが評価結果かチェック"""
        
        processing_stage = document.metadata.get("processing_stage", "")
        return processing_stage in ["evaluation", "test_results", "contradiction_detection"]
    
    def _extract_metrics(self, document: Document) -> Dict[str, float]:
        """ドキュメントからメトリクスを抽出"""
        
        metrics = {}
        
        # メタデータからメトリクスを抽出
        metadata = document.metadata
        
        # 一般的なメトリクス
        if "overall_score" in metadata:
            metrics["overall_score"] = float(metadata["overall_score"])
        
        if "success_rate" in metadata:
            metrics["accuracy"] = float(metadata["success_rate"])
        
        if "processing_time" in metadata:
            metrics["response_time"] = float(metadata["processing_time"])
        
        if "average_confidence" in metadata:
            metrics["confidence"] = float(metadata["average_confidence"])
        
        # コンテンツからメトリクスを解析
        content = document.content
        content_metrics = self._parse_metrics_from_content(content)
        metrics.update(content_metrics)
        
        return metrics
    
    def _parse_metrics_from_content(self, content: str) -> Dict[str, float]:
        """コンテンツからメトリクスを解析"""
        
        metrics = {}
        lines = content.split('\n')
        
        for line in lines:
            line = line.strip()
            
            # 精度情報
            if "精度" in line and ":" in line:
                try:
                    value_str = line.split(':')[1].replace('%', '').strip()
                    metrics["accuracy"] = float(value_str) / 100 if '%' in line else float(value_str)
                except:
                    pass
            
            # F1スコア
            elif "F1スコア" in line and ":" in line:
                try:
                    value_str = line.split(':')[1].strip()
                    metrics["f1_score"] = float(value_str)
                except:
                    pass
            
            # 応答時間
            elif "応答時間" in line and ":" in line:
                try:
                    value_str = line.split(':')[1].replace('秒', '').strip()
                    metrics["response_time"] = float(value_str)
                except:
                    pass
            
            # 信頼度
            elif "信頼度" in line and ":" in line:
                try:
                    value_str = line.split(':')[1].strip()
                    metrics["confidence"] = float(value_str)
                except:
                    pass
            
            # 一貫性スコア
            elif "一貫性" in line and ":" in line:
                try:
                    value_str = line.split(':')[1].replace('%', '').strip()
                    metrics["consistency"] = float(value_str) / 100 if '%' in line else float(value_str)
                except:
                    pass
        
        return metrics
    
    def _generate_insights(self, metrics: Dict[str, float], document: Document) -> List[Insight]:
        """メトリクスからインサイトを生成"""
        
        insights = []
        
        # 閾値ベースのインサイト生成
        threshold_insights = self._generate_threshold_insights(metrics)
        insights.extend(threshold_insights)
        
        # トレンド分析インサイト
        if self.config.enable_trend_analysis and len(self.historical_data) > 1:
            trend_insights = self._generate_trend_insights(metrics)
            insights.extend(trend_insights)
        
        # 比較分析インサイト
        if self.config.enable_comparative_analysis:
            comparative_insights = self._generate_comparative_insights(metrics)
            insights.extend(comparative_insights)
        
        # 根本原因分析インサイト
        if self.config.enable_root_cause_analysis:
            root_cause_insights = self._generate_root_cause_insights(metrics, document)
            insights.extend(root_cause_insights)
        
        # 信頼度フィルタリング
        filtered_insights = [i for i in insights if i.confidence >= self.config.min_confidence_for_insight]
        
        return filtered_insights
    
    def _generate_threshold_insights(self, metrics: Dict[str, float]) -> List[Insight]:
        """閾値ベースのインサイトを生成"""
        
        insights = []
        
        for threshold in self.all_thresholds:
            metric_name = threshold.metric_name
            if metric_name not in metrics:
                continue
            
            metric_value = metrics[metric_name]
            insight = self._evaluate_threshold(metric_name, metric_value, threshold)
            
            if insight:
                insights.append(insight)
        
        return insights
    
    def _evaluate_threshold(self, metric_name: str, value: float, threshold: Threshold) -> Optional[Insight]:
        """個別閾値の評価"""
        
        severity = "info"
        title = ""
        description = ""
        recommendations = []
        
        is_greater_than = threshold.comparison_operator == "greater_than"
        
        # 重要度判定
        if threshold.critical_threshold is not None:
            if is_greater_than:
                if value < threshold.critical_threshold:
                    severity = "critical"
            else:
                if value > threshold.critical_threshold:
                    severity = "critical"
        
        if severity != "critical" and threshold.warning_threshold is not None:
            if is_greater_than:
                if value < threshold.warning_threshold:
                    severity = "high"
            else:
                if value > threshold.warning_threshold:
                    severity = "high"
        
        if severity not in ["critical", "high"] and threshold.target_threshold is not None:
            if is_greater_than:
                if value < threshold.target_threshold:
                    severity = "medium"
                else:
                    severity = "low"  # 目標達成
            else:
                if value > threshold.target_threshold:
                    severity = "medium"
                else:
                    severity = "low"  # 目標達成
        
        # インサイト内容の生成
        if severity == "critical":
            title = f"{metric_name}が臨界レベルに達しています"
            description = f"{metric_name}の値({value:.3f})が臨界閾値を下回っており、緊急対応が必要です。"
            recommendations = [
                "緊急対応チームを招集してください",
                "システムの根本的な見直しを実施してください",
                "一時的にサービスを停止することを検討してください"
            ]
        elif severity == "high":
            title = f"{metric_name}に重要な問題があります"
            description = f"{metric_name}の値({value:.3f})が警告レベルにあり、早急な改善が必要です。"
            recommendations = [
                "改善計画を立案してください",
                "関連する設定やパラメータを見直してください",
                "監視を強化してください"
            ]
        elif severity == "medium":
            title = f"{metric_name}の改善余地があります"
            description = f"{metric_name}の値({value:.3f})は許容範囲内ですが、目標値に到達していません。"
            recommendations = [
                "最適化の機会を探してください",
                "ベストプラクティスを適用してください"
            ]
        elif severity == "low":
            title = f"{metric_name}は良好です"
            description = f"{metric_name}の値({value:.3f})は目標を達成しています。"
            recommendations = [
                "現在の設定を維持してください",
                "他のメトリクスの改善に集中してください"
            ]
        
        if severity in ["critical", "high", "medium"]:  # info レベルはスキップ
            insight = Insight(
                id=f"threshold_{metric_name}_{severity}",
                insight_type=self._classify_insight_type(metric_name),
                title=title,
                description=description,
                severity=severity,
                confidence=0.9,  # 閾値ベースは高信頼度
                affected_metrics=[metric_name],
                recommendations=recommendations,
                supporting_data={
                    "current_value": value,
                    "threshold_config": threshold.dict(),
                    "target_met": severity == "low"
                }
            )
            return insight
        
        return None
    
    def _classify_insight_type(self, metric_name: str) -> InsightType:
        """メトリクス名からインサイトタイプを分類"""
        
        type_mapping = {
            "accuracy": InsightType.QUALITY,
            "precision": InsightType.QUALITY,
            "recall": InsightType.QUALITY,
            "f1_score": InsightType.QUALITY,
            "response_time": InsightType.PERFORMANCE,
            "processing_time": InsightType.EFFICIENCY,
            "confidence": InsightType.RELIABILITY,
            "consistency": InsightType.CONSISTENCY,
            "throughput": InsightType.SCALABILITY,
            "error_rate": InsightType.RELIABILITY
        }
        
        return type_mapping.get(metric_name, InsightType.PERFORMANCE)
    
    def _generate_trend_insights(self, current_metrics: Dict[str, float]) -> List[Insight]:
        """トレンド分析インサイトを生成"""
        
        insights = []
        
        if len(self.historical_data) < 2:
            return insights
        
        # 過去のデータと比較
        prev_metrics = self.historical_data[-2]["metrics"]
        
        for metric_name, current_value in current_metrics.items():
            if metric_name in prev_metrics:
                prev_value = prev_metrics[metric_name]
                change_rate = (current_value - prev_value) / prev_value if prev_value != 0 else 0
                
                # 有意な変化があった場合のみインサイト生成
                if abs(change_rate) > 0.1:  # 10%以上の変化
                    insight = self._create_trend_insight(metric_name, current_value, prev_value, change_rate)
                    if insight:
                        insights.append(insight)
        
        return insights
    
    def _create_trend_insight(self, metric_name: str, current: float, previous: float, change_rate: float) -> Optional[Insight]:
        """トレンドインサイトを作成"""
        
        is_improvement = self._is_improvement(metric_name, change_rate)
        direction = "改善" if is_improvement else "悪化"
        severity = "low" if is_improvement else ("high" if abs(change_rate) > 0.3 else "medium")
        
        title = f"{metric_name}が{direction}しています"
        description = f"{metric_name}が前回の{previous:.3f}から{current:.3f}に変化しました（{change_rate:+.1%}）。"
        
        recommendations = []
        if is_improvement:
            recommendations = [
                "良好な傾向を維持してください",
                "成功要因を分析して他の領域に応用してください"
            ]
        else:
            recommendations = [
                "悪化の原因を特定してください",
                "対策を講じて改善を図ってください",
                "モニタリングを強化してください"
            ]
        
        insight = Insight(
            id=f"trend_{metric_name}_{direction}",
            insight_type=self._classify_insight_type(metric_name),
            title=title,
            description=description,
            severity=severity,
            confidence=0.8,
            affected_metrics=[metric_name],
            recommendations=recommendations,
            supporting_data={
                "current_value": current,
                "previous_value": previous,
                "change_rate": change_rate,
                "is_improvement": is_improvement
            }
        )
        
        return insight
    
    def _is_improvement(self, metric_name: str, change_rate: float) -> bool:
        """変化が改善かどうかを判定"""
        
        # 値が大きい方が良いメトリクス
        higher_is_better = ["accuracy", "precision", "recall", "f1_score", "confidence", "consistency", "throughput"]
        
        # 値が小さい方が良いメトリクス
        lower_is_better = ["response_time", "processing_time", "error_rate"]
        
        if metric_name in higher_is_better:
            return change_rate > 0
        elif metric_name in lower_is_better:
            return change_rate < 0
        else:
            return change_rate > 0  # デフォルトは大きい方が良い
    
    def _generate_comparative_insights(self, metrics: Dict[str, float]) -> List[Insight]:
        """比較分析インサイトを生成"""
        
        insights = []
        
        # メトリクス間の相関分析
        if "accuracy" in metrics and "confidence" in metrics:
            accuracy = metrics["accuracy"]
            confidence = metrics["confidence"]
            
            # 精度と信頼度の不一致を検出
            if abs(accuracy - confidence) > 0.3:
                insight = Insight(
                    id="accuracy_confidence_mismatch",
                    insight_type=InsightType.RELIABILITY,
                    title="精度と信頼度に大きな差があります",
                    description=f"精度({accuracy:.3f})と信頼度({confidence:.3f})の間に大きな差があります。",
                    severity="medium",
                    confidence=0.7,
                    affected_metrics=["accuracy", "confidence"],
                    recommendations=[
                        "信頼度校正の見直しを検討してください",
                        "モデルの不確実性推定を改善してください"
                    ],
                    supporting_data={
                        "accuracy": accuracy,
                        "confidence": confidence,
                        "difference": abs(accuracy - confidence)
                    }
                )
                insights.append(insight)
        
        # 性能と品質のトレードオフ分析
        if "response_time" in metrics and "accuracy" in metrics:
            response_time = metrics["response_time"]
            accuracy = metrics["accuracy"]
            
            # 高精度だが遅い場合
            if accuracy > 0.8 and response_time > 2.0:
                insight = Insight(
                    id="accuracy_speed_tradeoff",
                    insight_type=InsightType.EFFICIENCY,
                    title="高精度ですが応答速度が遅いです",
                    description="高い精度を維持していますが、応答時間が目標を上回っています。",
                    severity="medium",
                    confidence=0.8,
                    affected_metrics=["accuracy", "response_time"],
                    recommendations=[
                        "キャッシュの導入を検討してください",
                        "モデルの軽量化を検討してください",
                        "並列処理の最適化を行ってください"
                    ]
                )
                insights.append(insight)
        
        return insights
    
    def _generate_root_cause_insights(self, metrics: Dict[str, float], document: Document) -> List[Insight]:
        """根本原因分析インサイトを生成"""
        
        insights = []
        
        # 低性能の根本原因を分析
        if metrics.get("accuracy", 1.0) < 0.6:
            # ドキュメントメタデータから原因を推測
            metadata = document.metadata
            
            potential_causes = []
            
            # データ品質問題
            if metadata.get("contradictions_found", 0) > 0:
                potential_causes.append("データの矛盾が検出されています")
            
            # 処理問題
            if metadata.get("errors_encountered", 0) > 0:
                potential_causes.append("処理エラーが発生しています")
            
            # カバレッジ問題
            if metrics.get("coverage", 1.0) < 0.5:
                potential_causes.append("テストカバレッジが不十分です")
            
            if potential_causes:
                insight = Insight(
                    id="low_performance_root_cause",
                    insight_type=InsightType.QUALITY,
                    title="低性能の根本原因が特定されました",
                    description="システムの低性能について、以下の根本原因が考えられます。",
                    severity="high",
                    confidence=0.7,
                    affected_metrics=["accuracy"],
                    recommendations=[
                        "データクリーニングを実施してください",
                        "処理パイプラインを見直してください",
                        "テストデータを拡充してください"
                    ],
                    supporting_data={
                        "potential_causes": potential_causes,
                        "evidence": metadata
                    }
                )
                insights.append(insight)
        
        return insights
    
    def _compute_health_score(self, metrics: Dict[str, float]) -> float:
        """システムヘルススコアを計算"""
        
        if not metrics:
            return 0.0
        
        # 重要メトリクスの重み
        weights = {
            "accuracy": 0.3,
            "response_time": 0.2,
            "confidence": 0.2,
            "consistency": 0.15,
            "f1_score": 0.15
        }
        
        score = 0.0
        total_weight = 0.0
        
        for metric_name, weight in weights.items():
            if metric_name in metrics:
                value = metrics[metric_name]
                
                # 正規化（0-1の範囲に）
                if metric_name == "response_time":
                    # 応答時間は小さい方が良い
                    normalized_value = max(0.0, 1.0 - min(value / 5.0, 1.0))
                else:
                    # その他は大きい方が良い
                    normalized_value = min(value, 1.0)
                
                score += normalized_value * weight
                total_weight += weight
        
        return score / total_weight if total_weight > 0 else 0.0
    
    def _format_insight_report(self, insights: List[Insight], metrics: Dict[str, float]) -> str:
        """インサイトレポートをフォーマット"""
        
        if self.config.report_format == "json":
            return self._format_json_report(insights, metrics)
        else:
            return self._format_markdown_report(insights, metrics)
    
    def _format_markdown_report(self, insights: List[Insight], metrics: Dict[str, float]) -> str:
        """Markdown形式のインサイトレポート"""
        
        lines = ["# システムインサイトレポート\n"]
        
        # エグゼクティブサマリー
        if self.config.include_executive_summary:
            lines.append("## 📋 エグゼクティブサマリー")
            
            health_score = self._compute_health_score(metrics)
            lines.append(f"**システムヘルススコア**: {health_score:.2f}/1.00")
            
            if health_score >= 0.8:
                lines.append("🟢 **状態**: 健全 - システムは良好に動作しています")
            elif health_score >= 0.6:
                lines.append("🟡 **状態**: 注意 - いくつかの改善領域があります")
            elif health_score >= 0.4:
                lines.append("🟠 **状態**: 警告 - 重要な問題があります")
            else:
                lines.append("🔴 **状態**: 危険 - 緊急対応が必要です")
            
            # 重要インサイトのサマリー
            critical_insights = [i for i in insights if i.severity == "critical"]
            high_insights = [i for i in insights if i.severity == "high"]
            
            if critical_insights:
                lines.append(f"⚠️ **緊急対応が必要**: {len(critical_insights)}件の臨界的問題")
            if high_insights:
                lines.append(f"🔥 **早急な対応が必要**: {len(high_insights)}件の重要問題")
            
            lines.append("")
        
        # キーメトリクス
        lines.append("## 📊 キーメトリクス")
        for metric_name, value in metrics.items():
            if metric_name == "response_time":
                lines.append(f"- **{metric_name}**: {value:.3f}秒")
            elif "rate" in metric_name or "accuracy" in metric_name:
                lines.append(f"- **{metric_name}**: {value:.1%}")
            else:
                lines.append(f"- **{metric_name}**: {value:.3f}")
        lines.append("")
        
        # インサイト別セクション
        if insights:
            lines.append("## 🔍 インサイト")
            
            # 重要度別に分類
            severity_order = ["critical", "high", "medium", "low"]
            severity_icons = {
                "critical": "🚨",
                "high": "⚠️",
                "medium": "📈",
                "low": "✅"
            }
            
            for severity in severity_order:
                severity_insights = [i for i in insights if i.severity == severity]
                
                if severity_insights:
                    lines.append(f"### {severity_icons[severity]} {severity.upper()} 重要度")
                    
                    for insight in severity_insights:
                        lines.append(f"#### {insight.title}")
                        lines.append(f"**カテゴリ**: {insight.insight_type.value}")
                        lines.append(f"**信頼度**: {insight.confidence:.1%}")
                        lines.append(f"**説明**: {insight.description}")
                        
                        if insight.recommendations:
                            lines.append("**推奨事項**:")
                            for rec in insight.recommendations:
                                lines.append(f"- {rec}")
                        
                        lines.append("")
        
        # アクションアイテム
        if self.config.include_action_items and insights:
            lines.append("## 🎯 アクションアイテム")
            
            # 優先度順にアクションアイテムを整理
            critical_actions = []
            high_actions = []
            medium_actions = []
            
            for insight in insights:
                for rec in insight.recommendations:
                    action_item = f"{rec} (関連: {insight.title})"
                    
                    if insight.severity == "critical":
                        critical_actions.append(action_item)
                    elif insight.severity == "high":
                        high_actions.append(action_item)
                    else:
                        medium_actions.append(action_item)
            
            if critical_actions:
                lines.append("### 🚨 緊急対応（24時間以内）")
                for action in critical_actions[:5]:  # 上位5件
                    lines.append(f"1. {action}")
                lines.append("")
            
            if high_actions:
                lines.append("### ⚠️ 重要対応（1週間以内）")
                for action in high_actions[:5]:  # 上位5件
                    lines.append(f"1. {action}")
                lines.append("")
            
            if medium_actions:
                lines.append("### 📈 改善機会（1か月以内）")
                for action in medium_actions[:5]:  # 上位5件
                    lines.append(f"1. {action}")
        
        return "\n".join(lines)
    
    def _format_json_report(self, insights: List[Insight], metrics: Dict[str, float]) -> str:
        """JSON形式のインサイトレポート"""
        
        report = {
            "timestamp": datetime.now().isoformat(),
            "health_score": self._compute_health_score(metrics),
            "metrics": metrics,
            "insights": [insight.dict() for insight in insights],
            "summary": {
                "total_insights": len(insights),
                "critical_count": len([i for i in insights if i.severity == "critical"]),
                "high_count": len([i for i in insights if i.severity == "high"]),
                "medium_count": len([i for i in insights if i.severity == "medium"]),
                "low_count": len([i for i in insights if i.severity == "low"])
            }
        }
        
        return json.dumps(report, ensure_ascii=False, indent=2)
    
    def get_insight_summary(self) -> Dict[str, Any]:
        """インサイトサマリーを取得"""
        
        if not self.generated_insights:
            return {"message": "No insights generated yet"}
        
        severity_counts = {}
        type_counts = {}
        
        for insight in self.generated_insights:
            # 重要度別カウント
            severity = insight.severity
            severity_counts[severity] = severity_counts.get(severity, 0) + 1
            
            # タイプ別カウント
            insight_type = insight.insight_type.value
            type_counts[insight_type] = type_counts.get(insight_type, 0) + 1
        
        avg_confidence = statistics.mean([i.confidence for i in self.generated_insights])
        
        return {
            "total_insights": len(self.generated_insights),
            "severity_distribution": severity_counts,
            "type_distribution": type_counts,
            "average_confidence": avg_confidence,
            "recommendations_count": sum(len(i.recommendations) for i in self.generated_insights)
        }


"""
Date-based file filter
日付ベースのファイルフィルター
"""

from pathlib import Path
from datetime import datetime, date
from typing import Optional
from refinire_rag.loader.filters.base_filter import BaseFilter
from refinire_rag.loader.models.file_info import FileInfo


class DateFilter(BaseFilter):
    """
    Filter files based on modification date
    更新日時に基づいてファイルをフィルタリング
    
    This filter can include files modified after a certain date,
    before a certain date, or within a date range.
    このフィルターは特定の日付以降に更新されたファイル、
    特定の日付以前に更新されたファイル、または日付範囲内の
    ファイルを含めることができます。
    """
    
    def __init__(self, 
                 modified_after: Optional[datetime] = None,
                 modified_before: Optional[datetime] = None):
        """
        Initialize the date filter
        日付フィルターを初期化
        
        Args:
            modified_after: Include files modified after this date
            modified_before: Include files modified before this date
            modified_after: この日付以降に更新されたファイルを含める
            modified_before: この日付以前に更新されたファイルを含める
            
        Note:
            At least one of modified_after or modified_before must be specified.
            Both can be specified to create a date range.
            modified_afterまたはmodified_beforeの少なくとも一方を指定する必要があります。
            両方を指定して日付範囲を作成することもできます。
        """
        if modified_after is None and modified_before is None:
            raise ValueError("At least one of modified_after or modified_before must be specified")
        
        self.modified_after = modified_after
        self.modified_before = modified_before
        
        # Validate date range
        # 日付範囲を検証
        if (self.modified_after is not None and 
            self.modified_before is not None and 
            self.modified_after >= self.modified_before):
            raise ValueError("modified_after must be before modified_before")
    
    @classmethod
    def modified_since(cls, since_date: datetime) -> 'DateFilter':
        """
        Create filter for files modified since a specific date
        特定の日付以降に更新されたファイル用のフィルターを作成
        
        Args:
            since_date: Include files modified after this date
            since_date: この日付以降に更新されたファイルを含める
            
        Returns:
            DateFilter instance
            DateFilterインスタンス
        """
        return cls(modified_after=since_date)
    
    @classmethod
    def modified_until(cls, until_date: datetime) -> 'DateFilter':
        """
        Create filter for files modified until a specific date
        特定の日付まで更新されたファイル用のフィルターを作成
        
        Args:
            until_date: Include files modified before this date
            until_date: この日付以前に更新されたファイルを含める
            
        Returns:
            DateFilter instance
            DateFilterインスタンス
        """
        return cls(modified_before=until_date)
    
    @classmethod
    def modified_between(cls, start_date: datetime, end_date: datetime) -> 'DateFilter':
        """
        Create filter for files modified within a date range
        日付範囲内で更新されたファイル用のフィルターを作成
        
        Args:
            start_date: Include files modified after this date
            end_date: Include files modified before this date
            start_date: この日付以降に更新されたファイルを含める
            end_date: この日付以前に更新されたファイルを含める
            
        Returns:
            DateFilter instance
            DateFilterインスタンス
        """
        return cls(modified_after=start_date, modified_before=end_date)
    
    def should_include(self, file_path: Path, file_info: FileInfo = None) -> bool:
        """
        Determine if file should be included based on modification date
        更新日時に基づいてファイルが含まれるべきかどうかを判定
        
        Args:
            file_path: Path to the file being evaluated
            file_info: Optional FileInfo object with modification time
            file_path: 評価対象のファイルパス
            file_info: 更新時刻を含むオプションのFileInfoオブジェクト
            
        Returns:
            True if file should be included, False otherwise
            ファイルが含まれるべき場合True、そうでなければFalse
        """
        # Get modification time from FileInfo if available, otherwise from file
        # FileInfoから更新時刻を取得、利用できない場合はファイルから取得
        if file_info is not None:
            modified_time = file_info.modified_at
        else:
            try:
                stat = file_path.stat()
                modified_time = datetime.fromtimestamp(stat.st_mtime)
            except (OSError, IOError):
                # If file cannot be accessed, exclude it
                # ファイルにアクセスできない場合は除外
                return False
        
        # Check against after date
        # 以降日付をチェック
        if self.modified_after is not None and modified_time <= self.modified_after:
            return False
        
        # Check against before date
        # 以前日付をチェック
        if self.modified_before is not None and modified_time >= self.modified_before:
            return False
        
        return True
    
    def get_filter_name(self) -> str:
        """
        Get descriptive name for this filter
        このフィルターの説明的な名前を取得
        """
        if self.modified_after is not None and self.modified_before is not None:
            after_str = self.modified_after.strftime("%Y-%m-%d")
            before_str = self.modified_before.strftime("%Y-%m-%d")
            return f"DateFilter(between {after_str} and {before_str})"
        elif self.modified_after is not None:
            after_str = self.modified_after.strftime("%Y-%m-%d")
            return f"DateFilter(after {after_str})"
        else:
            before_str = self.modified_before.strftime("%Y-%m-%d")
            return f"DateFilter(before {before_str})"
    
    def get_description(self) -> str:
        """
        Get detailed description of this filter
        このフィルターの詳細説明を取得
        """
        if self.modified_after is not None and self.modified_before is not None:
            after_str = self.modified_after.strftime("%Y-%m-%d %H:%M:%S")
            before_str = self.modified_before.strftime("%Y-%m-%d %H:%M:%S")
            return f"Include files modified between {after_str} and {before_str}"
        elif self.modified_after is not None:
            after_str = self.modified_after.strftime("%Y-%m-%d %H:%M:%S")
            return f"Include files modified after {after_str}"
        else:
            before_str = self.modified_before.strftime("%Y-%m-%d %H:%M:%S")
            return f"Include files modified before {before_str}"
    
    def __repr__(self) -> str:
        """
        Developer representation of the filter
        フィルターの開発者向け表現
        """
        return f"DateFilter(modified_after={self.modified_after}, modified_before={self.modified_before})"


"""
FilterConfig model for managing file filters
ファイルフィルター管理用のFilterConfigモデル
"""

from dataclasses import dataclass, field
from typing import List, TYPE_CHECKING

# Forward declarations to avoid circular imports
# 循環インポートを避けるための前方宣言
if TYPE_CHECKING:
    from refinire_rag.loader.filters.base_filter import BaseFilter
    from refinire_rag.loader.filters.extension_filter import ExtensionFilter
    from refinire_rag.loader.filters.date_filter import DateFilter
    from refinire_rag.loader.filters.path_filter import PathFilter
    from refinire_rag.loader.filters.size_filter import SizeFilter

@dataclass
class FilterConfig:
    """
    Configuration container for file filters
    ファイルフィルターの設定コンテナ
    
    This class manages the collection of filters to be applied
    during directory scanning. Filters are used only for inclusion/exclusion
    decisions and do not add metadata to documents.
    このクラスは、ディレクトリスキャン中に適用される
    フィルターのコレクションを管理します。フィルターは包含/除外の
    判定にのみ使用され、文書にメタデータを追加しません。
    """
    extension_filter: 'ExtensionFilter' = None
    date_filter: 'DateFilter' = None
    path_filter: 'PathFilter' = None
    size_filter: 'SizeFilter' = None
    custom_filters: List['BaseFilter'] = field(default_factory=list)
    
    def get_all_filters(self) -> List['BaseFilter']:
        """
        Get all configured filters as a list
        設定されたすべてのフィルターをリストとして取得
        
        Returns:
            List of all active filters
            すべてのアクティブなフィルターのリスト
        """
        filters = []
        
        if self.extension_filter is not None:
            filters.append(self.extension_filter)
        if self.date_filter is not None:
            filters.append(self.date_filter)
        if self.path_filter is not None:
            filters.append(self.path_filter)
        if self.size_filter is not None:
            filters.append(self.size_filter)
        
        filters.extend(self.custom_filters)
        return filters
    
    def add_custom_filter(self, filter_instance: 'BaseFilter'):
        """
        Add a custom filter to the configuration
        設定にカスタムフィルターを追加
        
        Args:
            filter_instance: Custom filter instance
            filter_instance: カスタムフィルターインスタンス
        """
        self.custom_filters.append(filter_instance)
    
    def has_filters(self) -> bool:
        """
        Check if any filters are configured
        フィルターが設定されているかどうかをチェック
        
        Returns:
            True if at least one filter is configured
            少なくとも1つのフィルターが設定されている場合True
        """
        return len(self.get_all_filters()) > 0
    
    def get_filter_summary(self) -> dict:
        """
        Get a summary of configured filters
        設定されたフィルターのサマリーを取得
        
        Returns:
            Dictionary with filter information
            フィルター情報を含む辞書
        """
        all_filters = self.get_all_filters()
        
        return {
            'total_filters': len(all_filters),
            'has_extension_filter': self.extension_filter is not None,
            'has_date_filter': self.date_filter is not None,
            'has_path_filter': self.path_filter is not None,
            'has_size_filter': self.size_filter is not None,
            'custom_filters_count': len(self.custom_filters),
            'filter_names': [filter.get_filter_name() for filter in all_filters] if hasattr(all_filters[0], 'get_filter_name') and all_filters else []
        }
    
    def __str__(self) -> str:
        """
        String representation of the filter configuration
        フィルター設定の文字列表現
        """
        summary = self.get_filter_summary()
        return f"FilterConfig(total_filters={summary['total_filters']})"